import abc
import json
import re
from copy import deepcopy

import numpy as np

try:
    # noinspection PyProtectedMember
    from collections import Iterable
except ImportError:
    from typing import Iterable

from typing import Dict
from jsonschema import ValidationError

from .. params import conf
from .. params import bonds
from .. params.bonds import (
    ALLOWED_CONSTRAINT_METHODS,
    ALLOWED_CONSTRAINT_CHARS,
    TRANSFORM_TO_METHODS)
from ..params.transform import (
    BinaryInitialProperties,
    StarInitialProperties,
    SpotInitialProperties,
    PulsationModeInitialProperties,
    NuisanceInitialProperties
)
from ... import (
    units as u,
    utils,
    settings
)
from ... base.error import InitialParamsError
from ... utils import is_empty
from ... logger import getLogger
from ... binary_system.utils import calculate_sma_estimate

logger = getLogger('analytics.params.parameters')


def deflate_phenomena(flatten):
    """
    Conversion of model parameters concerning spots or pulsations from flat format
    (eg. 'primary@spots@spot1@longitude`) into the standard (nested) JSON format.

    :param flatten: Dict[str, Union[float, bool]]; spot or pulsation mode parameters in flat format
    :return: Dict;
    """
    result = dict()
    for phenom_uid, phenom_meta in flatten.items():
        _, _, label, param = str(phenom_uid).split(conf.PARAM_PARSER)
        if label not in result:
            result[label] = dict(label=label)
        result[label][param] = phenom_meta
    return result


def deserialize_result(result_dict: Dict) -> Dict:
    """
    Function converts dictionary of results in user format JSON to flat format.

    :param result_dict: Dict; standard input format::

        >>> result_dict = {
        >>>     'system': {
        >>>         'inclination': {},
        >>>         # ...
        >>>     },
        >>>     'primary': {
        >>>         't_eff': {},
        >>>         'spots': [
        >>>             {
        >>>                 'label': 'spot1'
        >>>                 'longitude': {},
        >>>                 ...
        >>>             },
        >>>             ...
        >>>         ]
        >>>     }
        >>>     ...
        >>> }

    :return: Dict; model parameters in flat format::

        >>> {
        >>>     'system@inclination': {},
        >>>     'primary@t_eff': {},
        >>>     'primary@spots@spot1@longitude': {},
        >>>     ...
        >>> }

    """
    data = {}

    if 'r_quared' in result_dict:
        data['r_squared'] = result_dict['r_squared']

    for system_slot in BinaryInitialParameters.__slots__:

        if system_slot in result_dict['system']:
            system_prop = result_dict['system'][system_slot]
            data.update({f'system{conf.PARAM_PARSER}{system_slot}': system_prop})

        if system_slot == conf.NUISANCE_PARSER and system_slot in result_dict:
            for nuisance_slot in NuisanceInitialPrameters.__slots__:
                if nuisance_slot in result_dict[conf.NUISANCE_PARSER] and conf.NUISANCE_PARSER in result_dict:
                    nuisance_prop = result_dict[conf.NUISANCE_PARSER][nuisance_slot]
                    data.update({f'{conf.NUISANCE_PARSER}{conf.PARAM_PARSER}{nuisance_slot}': nuisance_prop})

        if system_slot in ['primary', 'secondary'] and system_slot in result_dict:
            component_prop = result_dict[system_slot]

            for component_slot in StarInitialParameters.__slots__:
                if component_slot not in ['spots', 'pulsations'] and component_slot in result_dict[system_slot]:
                    data.update({f'{system_slot}@{component_slot}': component_prop[component_slot]})

                elif component_slot in ['spots', 'pulsations'] and component_slot in result_dict[system_slot]:
                    for phenom in component_prop[component_slot]:
                        for phenom_key, phenom_value in phenom.items():
                            if phenom_key not in ['label']:
                                data.update({
                                    f'{system_slot}@{component_slot[:-1]}@{phenom["label"]}@{phenom_key}': phenom_value
                                })
    return data


def serialize_result(result_dict: Dict) -> Dict:
    """
    Function converts dictionary of fit parameters in flat format back to user format.

    :param result_dict: Dict; fit parameters in flat format::

        >>> {
        >>>     'system@inclination': {},
        >>>     'primary@t_eff': {},
        >>>     'primary@spots@spot1@longitude': {},
        >>>     ...
        >>> }

    :return: Dict; standard (nested) fit parameter JSON::

        >>> retval = {
        >>>     'system': {
        >>>         'inclination': {},
        >>>         # ...
        >>>     },
        >>>     'primary': {
        >>>         't_eff': {},
        >>>         'spots': [
        >>>             {
        >>>                 'label': 'spot1'
        >>>                 'longitude': {},
        >>>                 ...
        >>>             },
        >>>             ...
        >>>         ]
        >>>     }
        >>>     ...
        >>> }

    """

    ret_dict = dict()

    for param, value in result_dict.items():
        if param == 'r_squared':
            ret_dict['r_squared'] = result_dict['r_squared']
            continue

        identificators = param.split(conf.PARAM_PARSER)

        if identificators[0] not in ret_dict:
            ret_dict[identificators[0]] = dict()

        if str(identificators[1]) in conf.COMPOSITE_FLAT_PARAMS:
            identificators[1] = f'{identificators[1]}s'

        if identificators[1] not in ret_dict[identificators[0]]:
            ret_dict[identificators[0]][identificators[1]] = dict()

        if str(identificators[0]).startswith('system'):
            ret_dict['system'].update({
                identificators[1]: value
            })
            continue

        if str(identificators[1]) not in ['spots', 'pulsations']:
            ret_dict[identificators[0]].update({
                identificators[1]: value
            })
        else:
            if str(identificators[2]) not in ret_dict[identificators[0]][identificators[1]]:
                ret_dict[identificators[0]][identificators[1]][identificators[2]] = dict()

            ret_dict[identificators[0]][identificators[1]][identificators[2]].update({
                "label": identificators[2],
                identificators[3]: value
            })

    # renormalize spots and pulastions if presented
    for component in settings.BINARY_COUNTERPARTS:
        if component in ret_dict:
            for phenomena in ['spots', 'pulsations']:
                if phenomena in ret_dict[component]:
                    ret_dict[component][phenomena] = [value for value in ret_dict[component][phenomena].values()]

    return ret_dict


def extend_json_with_atm_params(params, atmosphere_model=None, limb_darkening_coefficients=None):
    """
    Function will extend the initialization JSON with custom atmosphere-related parameters
    (atmosphere, limb_darkening_coefficients).

    :param params: Dict; flattened json used to initialize the system eg::

        >>> {
        >>>     'system@mass_ratio': 1.0,
        >>>         ...
        >>>     'primary@teff': 80000,
        >>>         ...
        >>>     'secondary@teff': 5000,
        >>>         ...
        >>>     },
        >>> }

    :param atmosphere_model: dict; desired atmosphere `atmosphere` models::

        >>> {
        >>>     'primary': 'bb',
        >>>     'secondary': 'ck04'
        >>> }

    :param limb_darkening_coefficients: dict; custom limb-darkening coefficients::

        >>> {
        >>>     'primary': {
        >>>         'bolometric': [0.5, 0.5],  # for logarithmic and square_root law
        >>>         'TESS': [0.68, 0.32]
        >>>     },
        >>>     'secondary': ...
        >>> }

    :return: dict; updated parameter json, eg::

        >>> {
        >>>     'system@mass_ratio': 1.0,
        >>>         ...
        >>>     'primary@teff': 80000,
        >>>     'primary@atmosphere': 'bb',
        >>>     'primary@limb_darkening_coefficients': {
        >>>         'bolometric': [0.5, 0.5],
        >>>         'TESS': [0.68, 0.32]
        >>>      }
        >>>         ...
        >>>     'secondary@teff': 5000,
        >>>     'secondary@atmosphere': 'ck04',
        >>>         ...
        >>>     },
        >>> }

    """
    var_names = ['atmosphere', 'limb_darkening_coefficients']
    for ii, atmosphere_param in enumerate([atmosphere_model, limb_darkening_coefficients]):
        if atmosphere_param is None:
            continue

        component_list = ','.join(list(params.keys()))
        for component, atm_param in atmosphere_param.items():
            if not atm_param:
                continue
            if component in component_list:
                params[f'{component}{conf.PARAM_PARSER}{var_names[ii]}'] = atm_param
            else:
                raise ValueError(f'Component {component} does not figure in your fit parameters JSON. Make sure that '
                                 f'your `{var_names[ii]}` contain `primary` and `secondary` components in case of '
                                 f'binary system and `star` in case of a single star.')

    return params


def check_initial_param_validity(x0: Dict[str, 'InitialParameter'], all_fit_params, mandatory_fit_params):
    """
    Checking if initial system parameters including surface features (spots and pulsations) are containing all
    necessary values and no invalid entries.

    :param x0: Dict[str, 'InitialParameter']; dictionary of initial parameters
    :param all_fit_params: List; list of all valid system parameters (spot and pulsation parameters excluded)
    :param mandatory_fit_params: List; list of mandatory system parameters (spot and pulsation parameters excluded)
    :return:
    """
    param_names = {key: val.value for key, val in x0.items() if not re.search(r"|".join(conf.COMPOSITE_FLAT_PARAMS), key)}
    utils.invalid_param_checker(param_names, all_fit_params, 'x0')
    utils.check_missing_params(mandatory_fit_params, param_names, 'x0')

    # checking validity of spot parameters and pulsations parameters
    params_map = {'spot': conf.SPOTS_PARAMETERS, 'pulsation': conf.PULSATIONS_PARAMETERS}
    for phenom in params_map:

        _ = {key: value for key, value in x0.items() if re.search(phenom, key)}
        _ = deflate_phenomena(_)

        for label, meta in _.items():
            meta = meta.copy()
            meta.pop('label', None)

            utils.invalid_param_checker(meta.keys(), params_map[phenom], label)
            utils.check_missing_params(params_map[phenom], meta, label)


def xs_reducer(xs):
    """
    Convert phases `xs` to a single list of photometric phases and produce an inverse mapping which enables to
    reconstruct the passband- (or component-)specific photometric phases from the resulting array in case of
    light curves (radial velocities).

    :param xs: Dict[str, numpy.array]; phases defined for each passband or ceomponent::

        {<passband>: <phases>} for light curves
        {<component>: <phases>} for radial velocities

    :return: Tuple; (numpy.array, Dict[str, List[int]]);
    """
    # this most likely cannot work corretly in python < 3.6, since dicts are not ordered in lower python versions
    x = np.hstack(list(xs.values())).flatten()
    y = np.arange(len(x)).tolist()
    reverse_dict = dict()
    for xs_key, phases in xs.items():
        reverse_dict[xs_key] = y[:len(phases)]
        del (y[:len(phases)])

    xs_reduced, inverse = np.unique(x, return_inverse=True)
    reverse = {band: inverse[indices] for band, indices in reverse_dict.items()}
    return xs_reduced, reverse


def renormalize_value(val, _min, _max):
    """
    Convert normalized value `val` to the actual value of the parameter value from interval specific
    for given parameter defined my `_min` and `_max`. Inverse function to `normalize_value`.

    :param val: float; eg. 0.5
    :param _min: float; bottom normalization value, eg. 5000 (K)
    :param _max: float; top normalization value, eg. 7000 (K)
    :return: float; eg. 6000 (K)
    """
    return (val * (_max - _min)) + _min


def normalize_value(val, _min, _max):
    """
    Normalize acutal value of the parameter `val` to value from interval (0, 1) based on `_min` and `_max`.

    :param val: float; eg. 6000 (K)
    :param _min: float; bottom normalization value, eg. 5000 (K)
    :param _max: float; top normalization value, eg. 7000 (K)
    :return: float; eg. 0.5
    """
    return (val - _min) / (_max - _min)


def vector_renormalizer(vector, properties, normalization):
    """
    Convert array of normalized parameters to their actual values according to the normalization boundaries.

    :param vector: List[float]; iterable of normalized parameter values
    :param properties: Iterable[str]; parameter names corresponding to `vector`
    :param normalization: Dict[str, Tuple[float, float]]; normalization map
                          {parameter@name: (bottom boundary, upper boundary), ...}
    :return: List[float]; restored values of parameters
    """
    return [renormalize_value(value, *normalization[prop]) for value, prop in zip(vector, properties)]


def vector_normalizer(vector, properties, normalization):
    """
    Normalize `vector` values from actual values to values between (0, 1).

    :param vector: List[float]; iterable of parameter values
    :param properties: Iterable[str]; parameter names corresponding to `vector`
    :param normalization: Dict[str, Tuple[float, float]]; normalization map
                          {parameter@name: (bottom boundary, upper boundary), ...}
    :return: List[float]; normalized values of parameters
    """
    return [normalize_value(value, *normalization[prop]) for value, prop in zip(vector, properties)]


def prepare_properties_set(xn, properties, constrained, fixed):
    """
    This will prepare final kwargs for synthetic model evaluation.

    :param xn: numpy.array; initial vector
    :param properties: List; variable labels corresponding to `xn`
    :param constrained: Dict; constrained initial parameters
    :param fixed: Dict; fixed initial parameters
    :return: Dict[str, float]; flat model parameters {'param@name': value, }
    """
    kwargs = dict(zip(properties, xn))
    kwargs.update(constraints_evaluator(kwargs, constrained))
    fixed = {key: val.value if isinstance(val, InitialParameter) else val for key, val in fixed.items()}
    kwargs.update(fixed)
    return kwargs


def prepare_nuisance_properties_set(xn, properties, fixed):
    """
    This will extract nuisance parameter values used during MCMC sampling.

    :param xn: numpy.array; initial vector
    :param properties: List; variable labels
    :param fixed: Dict; fixed initial parameters
    :return: Dict[str, float]; values of nuisance parameters
    """
    kwargs = {key: item for item, key in zip(xn, properties) if conf.NUISANCE_PARSER in key}
    kwargs.update({key: val.value if isinstance(val, InitialParameter) else val for key, val in fixed.items()
                   if conf.NUISANCE_PARSER in key})
    return kwargs


def check_for_invalid_constraint(constrained, allowed_params):
    """
    Making sure that constrains after substitution of valid parameters for their current values does not contain
    any invalid model parameters.

    :param constrained: dict; {parameter@name: constraint}
    :param allowed_params: Union[list, dict_keys]; names of valid model parameters
    :return: None
    """
    for c_param, constraint in constrained.items():
        if conf.PARAM_PARSER in constraint:
            raise InitialParamsError(f'Your constraint for parameter {c_param} is currently looks: {constraint} and '
                                     f'contains non-variable parameter. Only following parameters can be used to '
                                     f'define a constrained parameter: {allowed_params}')


def constraints_evaluator(substitution: Dict, constrained: Dict) -> Dict:
    """
    Substitute variables in constraint with values and evaluate to number.

    :param substitution: Dict[str, Union[InitialParameter, float]]; non-fixed values
                                                                   (xn vector in dict form {label: param})
    :param constrained: Dict[str, Union[InitialParameter, float]]; constraint parameters
                                                                   in form {label: InitialParameter}
    :return: Dict[str, float];
    """
    if is_empty(constrained):
        return dict()

    if isinstance(list(substitution.values())[-1], InitialParameter):
        substitution = {key: val.value for key, val in substitution.items()}

    if isinstance(list(constrained.values())[-1], InitialParameter):
        constrained = {key: val.constraint for key, val in constrained.items()}
    
    numpy_methods = [f'bonds.{method}' for method in TRANSFORM_TO_METHODS]
    allowed_methods = bonds.ALLOWED_CONSTRAINT_METHODS
    constrained = constrained.copy()

    subst = {key: utils.str_repalce(val, substitution.keys(), substitution.values())
             for key, val in constrained.items()}
    numpy_callable = {key: utils.str_repalce(val, allowed_methods, numpy_methods) for key, val in subst.items()}

    # raising error if invalid parameter is used to define a constraint
    check_for_invalid_constraint(numpy_callable, substitution.keys())

    try:
        evaluated = {key:  eval(val) for key, val in numpy_callable.items()}
    except Exception as e:
        raise InitialParamsError(f'Invalid syntax or value in constraint, {str(e)}.')
    return evaluated


def extend_result_with_sma(fit_parameters):
    """
    This function extends the fitting parameter dictionary for semi-major axis parameter that in case of fit,
    based solely on LC, has to be fixed on some sensible value.

    :param fit_parameters: Dict; input fitting parameters
    :return: result dictionary (with sma autofilled)
    """
    if 'mass_ratio' not in fit_parameters['system'].keys():
        logger.debug('Binary system parameters are supplied in standard format where SMA does not figure. '
                     'Nothing to add.')
        return fit_parameters

    if 'semi_major_axis' in fit_parameters['system'].keys():
        logger.debug('Binary system parameters already contains `semi_major_axis` parameters.')
        return fit_parameters

    sma_estimate = []
    mid_g = 270  # m.s^-2, a reasonable estimate of surface g
    mass_ratio = fit_parameters['system']['mass_ratio']['value']
    period = fit_parameters['system']['period']['value']
    for component in settings.BINARY_COUNTERPARTS.keys():
        synchronicity = fit_parameters[component].get('synchronicity', {'value': 1.0})['value']
        potential = fit_parameters[component]['surface_potential']['value']
        sma_estimate.append(calculate_sma_estimate(mass_ratio, synchronicity, potential, period, component, mid_g))

    fit_parameters['system']['semi_major_axis'] = {
        'value': np.mean(sma_estimate),
        'fixed': True,
        'unit': u.solRad
    }
    return fit_parameters


class ParameterMeta(object):
    """
    Auxiliary handler for fit parameters that ensures that each parameter attribute is present in the results.
    """
    def __init__(self, **kwargs):
        self.unit = kwargs.get("unit")
        self.param = kwargs.get("param")
        self.property = self.param
        self.fixed = kwargs.get("fixed")
        self.constraint = kwargs.get("constraint")
        self.value = kwargs.get("value")
        self.min = kwargs.get("min")
        self.max = kwargs.get("max")
        self.sigma = kwargs.get("sigma")

    def to_dict(self):
        return dict(
            **{
                "value": self.value,
                "param": self.param,
                "min": self.min,
                "max": self.max,
                "unit": str(self.unit) if self.unit is not None else None,
                "fixed": self.fixed,
                "constraint": self.constraint,
                "sigma": self.sigma
            }
        )


class InitialParameter(object):
    """
    Object designed for storing attributes of fit parameter.

    :param transform_cls: callable; class used to transform user fit parameters into default units
    :param kwargs: Dict; fit parameter attributes
    :**kwargs**:
        * **param**: str; parameter name
        * :value: float; value assigned to the fit parameter
        * :unit: astropy.unit.Unit; unit assigned to the `value`
        * :fixed: bool; if True, the parameter value will be fixed during the fit procedure
        * :constraint: Union[None, str]; enables to constrain the parameter value to the different variable
                                         fit parameter using mathematical expression. Here is the example of
                                         the constraining the semi_major_axis parameter to the value of
                                         inclination::

                                            'constraint': '12.528 / sin(system@inclination)'

                                         Each fit parameter definition can contain either
                                         `fixed` or `constraint` attribute at the same time.
        * :min: Union[None, float]; minimum allowed value for the parameter
        * :max: Union[None, float]; maximum allowed value for the parameter, `max` > `min`
        * :sigma: Union[None, float]; standard deviation of the normal prior distribution for given parameter,
                                      usefull only in case of MCMC method

    """
    DEFAULT = {
        "param": None,
        "value": None,
        "fixed": None,
        "constraint": None,
        "min": None,
        "max": None,
        "unit": None,
        "sigma": None
    }

    def __init__(self, transform_cls, **kwargs):
        self.unit = kwargs.get("unit")
        self.param = kwargs.get("param")
        self.property = self.param
        self.fixed = kwargs.get("fixed")
        self.constraint = kwargs.get("constraint")
        self.value = None
        self.min = None
        self.max = None
        self.sigma = None

        # units transformaton
        self.unit = u.Unit(self.unit) if self.unit is not None else self.unit
        if self.unit is not None and self.unit is not u.dimensionless_unscaled and self.constraint is None:
            kwargs.update({
                "value": kwargs.get("value") * self.unit,
                "min": kwargs.get("min") * self.unit,
                "max": kwargs.get("max") * self.unit,
            })
            if kwargs.get("sigma") is not None:
                kwargs.update({"sigma": kwargs.get("sigma") * self.unit})

        if self.constraint is None:
            self.value = transform_cls.transform_input(**{self.param: kwargs.get("value")})[self.param]
            self.min = transform_cls.transform_input(**{self.param: kwargs.get("min")})[self.param]
            self.max = transform_cls.transform_input(**{self.param: kwargs.get("max")})[self.param]
            if kwargs.get("sigma") is not None:
                self.sigma = transform_cls.transform_input(**{self.param: kwargs.get("sigma")})[self.param]

        if self.fixed:
            self.min, self.max, self.sigma = None, None, None

        self.unit = conf.DEFAULT_FLOAT_UNITS[self.property]

    def copy(self):
        """Returns deep copy of the InitialParameter"""
        return deepcopy(self)

    def to_dict(self):
        """Transforms InitialParameter into dict format."""
        return dict(
            **{
                "value": self.value,
                "param": self.param,
                "min": self.min,
                "max": self.max,
                "unit": str(self.unit) if self.unit is not None else None,
                "sigma": self.sigma
            },
            **{"fixed": self.fixed} if self.fixed is not None else {},
            **{"constraint": self.constraint} if self.constraint is not None else {},
        )

    def __repr__(self):
        return json.dumps({
            "param": self.param,
            "value": self.value,
            "fixed": self.fixed,
            "constraint": self.constraint,
            "min": self.min,
            "max": self.max,
            "sigma": self.sigma,
            "unit": str(self.unit) if self.unit is not None else None
        }, indent=4)

    __str__ = __repr__


class InitialParameters(object, metaclass=abc.ABCMeta):
    """
    Abstract object for handling sets of initial fit parameters.
    """
    TRANSFORM_PROPERTIES_CLS = None
    DEFAULT_NORMALIZATION = None

    @property
    def slots_(self):
        if hasattr(self, '__slots__'):
            return getattr(self, '__slots__')
        return []

    def validity_check(self):
        """
        Function examines whether inputted definitions of initial fit parameters make sense.
        Verifying that InitialParameter.min <= InitialParameter.value <= InitialParameter.max.
        Verifying that InitialParameter is either:

            * fixed: `InitialParameter.fixed` = False, `InitialParameter.constraint` = None
            * variable: `InitialParameter.fixed` = True, `InitialParameter.constraint` = None
            * constrained: `InitialParameter.fixed` = None, `constraint` = `InitialParameter.valid expresion`

        """
        for slot in self.slots_:
            if not hasattr(self, str(slot)):
                continue
            prop = getattr(self, str(slot))

            if not isinstance(prop, InitialParameter):
                continue

            if prop.constraint is None and not prop.fixed:
                if not (prop.min <= prop.value <= prop.max):
                    raise InitialParamsError(f'Initial parameters in parameter `{prop.param}` are not valid. '
                                             f'Invalid bounds: {prop.min} <= {prop.value} <= {prop.max}')
            if prop.fixed is not None and prop.constraint is not None:
                raise InitialParamsError(f'It is not allowed for `{prop.param}` to contain '
                                         f'`fixed` and `constraint` parameter.')

    def init_parameter(self, parameter: str, items: Dict) -> InitialParameter:
        """
        Initialization of InitialParameter instance from dict definition.

        :param parameter: str; name of the fit parameter
        :param items: Dict; definition of the fit parameter {'value': , 'unit':, 'fixed':, ...}
        :return: InitialParameter;
        """
        _kwarg = InitialParameter.DEFAULT.copy()
        _kwarg.update(dict(param=parameter, **items))
        _kwarg.update(dict(min=items.get("min", self.DEFAULT_NORMALIZATION[parameter][0]),
                           max=items.get("max", self.DEFAULT_NORMALIZATION[parameter][1])))
        return InitialParameter(transform_cls=self.__class__.TRANSFORM_PROPERTIES_CLS, **_kwarg)


class SpotInitialParameters(InitialParameters):
    """
    Object for handling spot-related fit parameters.
    """
    __slots__ = ["longitude", "latitude", "angular_radius", "temperature_factor", "label"]

    TRANSFORM_PROPERTIES_CLS = SpotInitialProperties
    DEFAULT_NORMALIZATION = conf.DEFAULT_NORMALIZATION_SPOT

    def __init__(self, **kwargs):
        self.label = f'spot{conf.PARAM_PARSER}{kwargs.pop("label")}'
        for parameter, items in kwargs.items():
            value = self.init_parameter(parameter, items)
            setattr(self, parameter, value)
        self.validity_check()


class PulsationInitialParameters(InitialParameters):
    """
    Object for handling pulsation mode-related fit parameters.
    """
    __slots__ = ["l", "m", "amplitude", "frequency", "start_phase", "mode_axis_theta", "mode_axis_phi", "label"]

    TRANSFORM_PROPERTIES_CLS = PulsationModeInitialProperties
    DEFAULT_NORMALIZATION = conf.DEFAULT_NORMALIZATION_PULSATION

    def __init__(self, **kwargs):
        self.label = f'pulsation{conf.PARAM_PARSER}{kwargs.pop("label")}'
        self.mode_axis_phi = 0
        self.mode_axis_theta = 0

        for parameter, items in kwargs.items():
            value = self.init_parameter(parameter, items)
            setattr(self, parameter, value)
        self.validity_check()


class StarInitialParameters(InitialParameters):
    """
    Object for handling component-related fit parameters.
    """
    __slots__ = ["spots", "pulsations", "t_eff", "metallicity", "surface_potential",
                 "albedo", "gravity_darkening", "synchronicity", "mass"]

    TRANSFORM_PROPERTIES_CLS = StarInitialProperties
    DEFAULT_NORMALIZATION = conf.DEFAULT_NORMALIZATION_STAR

    def __init__(self, **kwargs):
        self.label = None

        spots = kwargs.pop('spots', [])
        pulsations = kwargs.pop('pulsations', [])

        for parameter, items in kwargs.items():
            value = self.init_parameter(parameter, items)
            setattr(self, parameter, value)

        if not is_empty(spots):
            spots = [SpotInitialParameters(**spot) for spot in spots]
            self.spots = spots

        if not is_empty(pulsations):
            pulsations = [PulsationInitialParameters(**pulsation) for pulsation in pulsations]
            self.pulsations = pulsations

        self.validity_check()


class NuisanceInitialPrameters(InitialParameters):
    """
    Object for handling auxiliary `nuisance` fit parameters used during the MCMC fit to produce a
    realistic estimations of confidence intervals.
    """
    __slots__ = ["ln_f"]

    TRANSFORM_PROPERTIES_CLS = NuisanceInitialProperties
    DEFAULT_NORMALIZATION = conf.DEFAULT_NORMALIZATION_NUISANCE

    def __init__(self, **kwargs):
        for parameter, items in kwargs.items():
            value = self.init_parameter(parameter, items)
            setattr(self, parameter, value)
        self.validity_check()


class BinaryInitialParameters(InitialParameters):
    """
    Object for handling binary system-related fit parameters.
    """
    __slots__ = ["primary", "secondary", "eccentricity", "argument_of_periastron",
                 "inclination", "gamma", "period", "mass_ratio", "asini", "semi_major_axis",
                 "additional_light", "phase_shift", "primary_minimum_time", conf.NUISANCE_PARSER]

    TRANSFORM_PROPERTIES_CLS = BinaryInitialProperties
    DEFAULT_NORMALIZATION = conf.DEFAULT_NORMALIZATION_SYSTEM

    def __init__(self, **kwargs):
        self._primary = kwargs.pop('primary', None)
        self._secondary = kwargs.pop('secondary', None)
        self._nuisance = kwargs.pop(conf.NUISANCE_PARSER, None)
        system = kwargs.pop('system')

        # nuisance params
        if self._nuisance is None:
            self._nuisance = {"ln_f": {"value": -20, "fixed": True}}
        self.nuisance = NuisanceInitialPrameters(**self._nuisance)

        # system params
        for parameter, items in system.items():
            value = self.init_parameter(parameter, items)
            setattr(self, parameter, value)

        # components params
        for component in settings.BINARY_COUNTERPARTS:
            props = getattr(self, f'_{component}')
            if is_empty(props):
                continue
            setattr(self, component, StarInitialParameters(**props))
            setattr(getattr(self, component), 'label', component)

        self.validity_check()
        self.unique_labels_validation()
        self._data = self.serialize_flat_set()
        self.validate_data()

    @property
    def data(self):
        """Returns a complete set of binary system model parameters
        compiled from BinaryInitialParameters instance in standard nested JSON format."""
        return self._data

    def __getitem__(self, item):
        iterable = True
        if not isinstance(item, Iterable):
            iterable = True
            item = [item]
        data = [self.data[_item] for _item in item]
        return data if iterable else data[-1]

    def to_flat_json(self):
        """
        Returns a complete set of binary system model parameters
        compiled from BinaryInitialParameters instance in flat JSON format.
        :return:
        """
        return [{key: val.to_dict()} for key, val in self.data.items()]

    def unique_labels_validation(self):
        """
        Making sure that labels of surface features (spots or pulsation modes have unique names.
        """
        def _test(_what):
            for component in settings.BINARY_COUNTERPARTS:
                if hasattr(self, component):
                    inst = getattr(self, component)
                    if hasattr(inst, _what):
                        phenom = getattr(inst, _what)
                        _all = [_phenom.label for _phenom in phenom]
                        if len(_all) != len(set(_all)):
                            raise InitialParamsError(f'It is not allowed to have multiple {_what} with same label.')
        _test('spots')
        _test('pulsations')

    def serialize_flat_set(self):
        """
        Returns a nested set of 'InitialParameter's in the same JSON format as fit results.

        :return: Dict[str, Union[Dict[...], InitialParameter]];
        """
        data = {}

        for system_slot in self.__slots__:

            if not hasattr(self, str(system_slot)):
                continue
            system_prop = getattr(self, str(system_slot))
            if isinstance(system_prop, InitialParameter):
                data.update({f'system{conf.PARAM_PARSER}{system_slot}': system_prop})
            elif isinstance(system_prop, NuisanceInitialPrameters):
                for nuisance_slot in system_prop.__slots__:
                    key = f'{conf.NUISANCE_PARSER}{conf.PARAM_PARSER}{nuisance_slot}'
                    data.update({key: getattr(system_prop, nuisance_slot)})
            elif isinstance(system_prop, StarInitialParameters):

                for component_slot in system_prop.__slots__:

                    if not hasattr(system_prop, str(component_slot)):
                        continue
                    component_prop = getattr(system_prop, str(component_slot))
                    if isinstance(component_prop, InitialParameter):

                        data.update({f'{getattr(system_prop, "label")}'
                                     f'{conf.PARAM_PARSER}{component_slot}': component_prop})

                    elif isinstance(component_prop, list):
                        for phenomena_prop in component_prop:
                            if isinstance(phenomena_prop, (PulsationInitialParameters, SpotInitialParameters)):

                                for phenomena_slot in phenomena_prop.__slots__:
                                    if not hasattr(phenomena_prop, str(phenomena_slot)):
                                        continue
                                    prop = getattr(phenomena_prop, str(phenomena_slot))

                                    if isinstance(prop, InitialParameter):
                                        data.update({f'{getattr(system_prop,"label")}'
                                                     f'{conf.PARAM_PARSER}{getattr(phenomena_prop,"label")}'
                                                     f'{conf.PARAM_PARSER}{phenomena_slot}': prop})

        return data

    def constraint_validator(self):
        """
        Validate constraints. Make sure there is no harmful code.
        Allowed methods used in constraints::

            'arcsin', 'arccos', 'arctan', 'log', 'sin', 'cos', 'tan', 'exp', 'degrees', 'radians'

        Allowed characters used in constraints::

            '(', ')', '+', '-', '*', '/', '.', 'e'

        :raise: elisa.base.error.ValidationError;
        """
        constrained = self.get_constrained(jsonify=False)
        substitution = self.get_substitution_dict()
        try:
            subst = {param:  utils.str_repalce(
                             utils.str_repalce(constraint.constraint, substitution.keys(), substitution.values()),
                             ALLOWED_CONSTRAINT_METHODS, [''] * len(ALLOWED_CONSTRAINT_METHODS)
                         ).replace(' ', '')
                     for param, constraint in constrained.items()}
        except KeyError:
            msg = f'It seems your constraint contain variable that cannot be resolved. ' \
                f'Make sure that linked constraint variable is not fixed or check for typos in variable name in ' \
                f'constraint expression.'
            raise ValidationError(msg)

        for key, val in subst.items():
            if not np.all(np.isin(list(val), ALLOWED_CONSTRAINT_CHARS)):
                msg = f'Constraint {key} contain forbidden characters. Allowed: {ALLOWED_CONSTRAINT_CHARS}'
                raise ValidationError(msg)

    def validate_data(self):
        """
        Validate whether fit parameters contain at least one variable parameter. Function also checks
        constrained parameters for validity.
        """
        # validate that at least one parameter is not fixed
        if len(self.get_fitable(jsonify=False)) == 0:
            raise ValidationError('There are no variable parameters to fit.')

        # constraint validation
        self.constraint_validator()

    def adjust_overcontact_potential(self, morphology):
        """
        If overcontact morphology is expected during the fit, this function constrains the secondary
        component potential to the primary component potential.

        :param morphology: str; over-contact or detached
        """
        if self.is_overcontact(morphology):
            update_surface_potential = self.primary.surface_potential.copy()
            update_surface_potential.fixed = None
            update_surface_potential.constraint = 'primary@surface_potential'
            self.secondary.surface_potential = update_surface_potential
            self.data['secondary@surface_potential'] = update_surface_potential

    def validate_lc_parameters(self, morphology):
        """
        Validate parameters for light curve fitting.
        """
        mandatory_fit_params = ['system@eccentricity', 'system@argument_of_periastron',
                                'system@period', 'system@inclination', 'system@period'] + \
                               [f'{component}@{param}'
                                for param in ['t_eff', 'surface_potential']
                                for component in settings.BINARY_COUNTERPARTS]

        optional_fit_params = ['system@semi_major_axis', 'system@primary_minimum_time', 'system@phase_shift',
                               'system@asini', 'system@mass_ratio', 'system@additional_light', 'nuisance@ln_f'] + \
                              [f'{component}@{param}'
                               for param in ['mass', 'synchronicity', 'metallicity', 'spots', 'pulsations',
                                             'gravity_darkening', 'albedo']
                               for component in settings.BINARY_COUNTERPARTS]

        all_fit_params = mandatory_fit_params + optional_fit_params
        utils.check_missing_kwargs(mandatory_fit_params, self.data, instance_of=self.__class__)
        check_initial_param_validity(self.data, all_fit_params, mandatory_fit_params)

        is_oc = self.is_overcontact(morphology)
        are_same = self.data['primary@surface_potential'].value == self.data['secondary@surface_potential'].value

        is_fixed_omega_1 = self.data['primary@surface_potential'].fixed or False
        is_fixed_omega_2 = self.data['secondary@surface_potential'].fixed or False

        any_fixed = is_fixed_omega_1 | is_fixed_omega_2
        all_fixed = is_fixed_omega_1 & is_fixed_omega_2

        if is_oc and all_fixed and are_same:
            return
        if is_oc and all_fixed and not are_same:
            msg = 'Different potential in over-contact morphology with all fixed (potential) value are not allowed.'
            raise ValidationError(msg)
        if is_oc and any_fixed:
            msg = 'Just one fixed potential in over-contact morphology is not allowed.'
            raise ValidationError(msg)

        # adjust constraint for secondary potential
        self.adjust_overcontact_potential(morphology)

    def validate_rv_parameters(self):
        """
        Validate parameters for radial velocities curve fitting.
        """
        mandatory_fit_params = ['system@eccentricity', 'system@argument_of_periastron', 'system@gamma']
        optional_fit_params = ['system@period', 'system@primary_minimum_time', 'primary@mass', 'secondary@mass',
                               'system@inclination', 'system@asini', 'system@mass_ratio', 'nuisance@ln_f']
        all_fit_params = mandatory_fit_params + optional_fit_params
        utils.check_missing_kwargs(mandatory_fit_params, self.data, instance_of=self.__class__)
        check_initial_param_validity(self.data, all_fit_params, mandatory_fit_params)

        # validate consistency of parameters (system has to be definable)
        has_primary_minimum_time, has_period = 'system@primary_minimum_time' in self.data, 'system@period' in self.data
        if has_primary_minimum_time:
            if not (has_primary_minimum_time and has_period):
                raise ValidationError("Input requires both, period and primary minimum time.")
        else:
            if not has_period:
                raise ValidationError("Input requires at least period.")

    def _get_kind_of(self, kind_of, jsonify=False):
        if jsonify:
            return {key: val.to_dict() for key, val in self.data.items() if getattr(val, kind_of)}
        return {key: val for key, val in self.data.items() if getattr(val, kind_of)}

    def get_fixed(self, jsonify=False):
        """
        Transform native form to `fixed` parameters only: {key: InitialParam, ...}
        :return: Dict[str, InitialParam];
        """
        return self._get_kind_of(kind_of="fixed", jsonify=jsonify)

    def get_constrained(self, jsonify=False):
        """
        Transform native form to `constrained` parameters only: {key: InitialParam, ...}
        :return: Dict[str, InitialParam];
        """
        return self._get_kind_of(kind_of="constraint", jsonify=jsonify)

    def get_fitable(self, jsonify=False):
        """
        Transform native form to `fitable` (variable) parameters only: {key: InitialParam, ...}
        :return: Dict[str, InitialParam];
        """
        if jsonify:
            return {key: val.to_dict() for key, val in self.data.items() if not val.constraint and not val.fixed}
        return {key: val for key, val in self.data.items() if not val.constraint and not val.fixed}

    def get_substitution_dict(self):
        """
        Return dictionary of fit parameter models containing parameter name and its
        value except for constrained parameters.

        :return: Dict[str, float]; {param@name}
        """
        return {key: val.value for key, val in self.data.items() if not val.constraint}

    def get_normalization_map(self):
        """
        Return normalization boundaries for given parmeter.
        :return: Tuple[float, float];
        """
        return {key: (val.min, val.max) for key, val in self.data.items()}

    @staticmethod
    def is_overcontact(morphology):
        """
        Checking if system morphology is over-contact.
        """
        return morphology in ['over-contact', 'overcontact']



import numpy as np
from .. base import error


def validate_period(period):
    if period <= 0:
        raise error.ValidationError("Period has to be > 0.")


def validate_pirmary_minimum_time(t0):
    if t0 <= 0:
        raise error.ValidationError("Primary minimum time has to be > 0.")


def adjust_phases(phases, centre=0.5):
    """
    shift phases to centre them on given value

    :param phases: Union[float, numpy.array];
    :param centre: float; centre around which phases will be calculated (+-0.5 around `centre` value)
    :return:
    """
    if isinstance(phases, list):
        phases = np.array(phases)
    shift = centre - 0.5
    return (phases - shift) % 1.0 + shift


def jd_to_phase(t0, period, jd, centre=0.5):
    """
    Convert JD time to phase

    :param centre: float; centre around which phases will be calculated (+-0.5 around `centre` value)
    :param t0: float; reference primary minimum time
    :param period: float; period of binary system
    :param jd: Union[float, numpy.array]; measurement JD times
    :return: Union[float, numpy.array]; phases
    """
    validate_period(period)
    validate_pirmary_minimum_time(t0)

    if isinstance(jd, list):
        jd = np.array(jd)
    shift = centre - 0.5
    return (((jd - t0) / period) - shift) % 1.0 + shift


def phase_to_jd(t0, period, phases):
    """
    Convert phase to JD time

    :param t0: float; reference primary minimum time
    :param period: float; period of binary system
    :param phases: Union[float, numpy.array]; phases
    :return: Union[float, numpy.array]; phases
    """
    validate_period(period)
    validate_pirmary_minimum_time(t0)

    if isinstance(phases, list):
        phases = np.array(phases)
    return (period * phases) + t0



import numpy as np

from copy import deepcopy
from . surface import (
    mesh,
    faces,
    gravity,
    temperature,
    pulsations
)
from .. base.container import (
    StarContainer,
    PositionContainer
)
from .. logger import getLogger

logger = getLogger("single_system.container")


class SinglePositionContainer(PositionContainer):
    """
    Object for handling models of a single star system at given orbital position.
    Use functions SinglePositionContainer.from_single_system or
    SingleSystem.build_container to correctly initialize this container.
    """
    def __init__(self, star: StarContainer, position, **properties):
        super().__init__(position=position)
        self._components = ['star']
        self.star = star

        # placeholder (set in loop below)
        self.rotation_period = np.nan

        for key, val in properties.items():
            setattr(self, key, val)

        # calculating a time that elapsed since t0
        self.set_time()

        # setting centre of mass
        self.set_com()

    def set_on_position_params(self, position):
        """
        Defining the orientation of the single system container with respect to the observer.

        :param position: elisa.const.Position;
        :return: SinglePositionContainer; container with set position
        """
        setattr(self, "position", position)
        return self

    def set_com(self):
        """
        Setting a centre of mass of the single star object.
        """
        setattr(self.star, 'com', np.array([0, 0, 0]))

    def set_time(self):
        """
        Calculating elapsed from `reference_time` for the instance of SinglePositionContainer.

        :return: float; time in seconds corresponding to the container
        """
        setattr(self, 'time', 86400 * self.rotation_period * self.position.phase)
        return getattr(self, 'time')

    @classmethod
    def from_single_system(cls, single_system, position):
        """
        Construct an SinglePositionContainer based on the instance of SingleSystem and position.

        :param single_system: elisa.single_system.system.SingleSystem;
        :param position: elisa.const.Position; named tuple containing information system orientation
                                               in space with respect to the observer.
        :return: SinglePositionContainer;
        """
        star = StarContainer.from_star_instance(single_system.star)
        return cls(star, position, **single_system.properties_serializer())

    def copy(self):
        """Returns a deep copy of the SinglePositionContainer."""
        return deepcopy(self)

    def has_spots(self):
        """Returns True if the star contains spots."""
        return self.star.has_spots()

    def has_pulsations(self):
        """Returns True if the star contains pulsations."""
        return self.star.has_pulsations()

    def build(self, build_pulsations=True, **kwargs):
        """
        Main method to build binary star system from parameters given on init of SingleStar.

        called following methods::

            - build_mesh
            - build_faces
            - build_velocities
            - build_surface_gravity
            - build_faces_orientation
            - correct_mesh
            - build_surface_areas
            - build_temperature_distribution

        :param kwargs:
        :param build_pulsations: bool;  if True, only equilibrium model is build
        :return: self;
        """
        self.build_surface()
        self.build_from_points()

        self.flat_it()
        if build_pulsations:
            self.build_pulsations()
        return self

    def build_pulsations(self):
        """
        Incorporating user-defined pulsation modes into the model.
        """
        self.build_harmonics()
        self.build_perturbations()

    def build_surface(self):
        """
        Building only clear surface. (points, faces, velocities)
        """
        self.build_mesh()
        self.build_faces()
        self.build_velocities()

    def build_from_points(self):
        """
        Build single system from present surface points

        :return: SingleSystemPosition;
        """
        self.build_surface_gravity()
        self.build_faces_orientation()
        self.correct_mesh()
        self.build_surface_areas()
        self.build_temperature_distribution()
        return self

    def build_mesh(self):
        """
        Build surface point mesh including spots.

        :return: elisa.single_system.container.SinglePositionContainer; container updated with point mesh
        """
        return mesh.build_mesh(self)

    def correct_mesh(self):
        """
        Correcting the underestimation of the surface due to the discretization.

        :return: elisa.single_system.container.SinglePositionContainer; container updated with corrected point mesh
        """
        return mesh.correct_mesh(self)

    def build_faces(self):
        """
        Function tessellates the stellar surface points into a set of triangles
        covering the star without gaps and overlaps.

        :return: elisa.single_system.container.SinglePositionContainer; container updated with faces
        """
        return faces.build_faces(self)

    def build_velocities(self):
        """
        Function calculates velocity vector for each face relative to the observer.

        :return: elisa.single_system.container.SinglePositionContainer; container updated with face velocities
        """
        return faces.build_velocities(self)

    def build_surface_areas(self):
        """
        Compute surface areas of all faces (spots included).

        :return: system; elisa.single_system.container.SinglePositionContainer; container updated with face
                                                                                (triangle) areas
        """
        return faces.compute_all_surface_areas(self)

    def build_faces_orientation(self):
        """
        Compute face orientation (normals) for each face.

        :return: elisa.single_system.container.SinglePositionContainer; container updated with correct normal vector
                                                                        orientation for each face
        """
        return faces.build_faces_orientation(self)

    def build_surface_gravity(self):
        """
        Function calculates gravity potential gradient magnitude (surface gravity) for each face.

        :return: elisa.single_system.container.SinglePositionContainer; container updated with surface gravity
                                                                        distribution
        """
        return gravity.build_surface_gravity(self)

    def build_temperature_distribution(self):
        """
        Function calculates temperature distribution on across all faces.

        :return: elisa.single_system.container.SinglePositionContainer; container updated with surface
                                                                        temperature distribution
        """
        return temperature.build_temperature_distribution(self)

    def build_harmonics(self):
        """
        Adds pre-calculated harmonics for the respective pulsation modes.

        :return: elisa.single_system.container.SinglePositionContainer; updated container
        """
        return pulsations.build_harmonics(self)

    def build_perturbations(self):
        """
        Function adds perturbation to the surface mesh due to pulsations.

        :return: elisa.single_system.container.SinglePositionContainer; container with introduced pulsations
        """
        return pulsations.build_perturbations(self)

    def _phase(self, phase):
        return phase if phase is not None else self.position.phase



import numpy as np

from ... pypex.base.conf import ROUND_PRECISION


def _line_side(p1, p2, a, b):
    """
    Idea is described in ``
    Help to determine, whether both point of line p lie on the same side of line ab or not.

    :param p1: numpy.array; firs point of line p
    :param p2: numpy.array; second point of line p
    :param a: numpy.array; first point of line p'
    :param b: numpy.array; second point of line p'
    :return: float
    """
    p1, p2, a, b = np.array(p1), np.array(p2), np.array(a), np.array(b)
    cp1 = np.cross(b - a, p1 - a)
    cp2 = np.cross(b - a, p2 - a)
    return np.dot(cp1, cp2)


# /* Check whether p1 and p2 lie on the same side of line ab */
# todo: reconsider to use edge definition instead of single point in arguments
def same_side(p1, p2, a, b):
    """
    Idea is described in ``
    Determine, whether both point of line p lie on the same side of line ab or not.

    :param p1: numpy.array; firs point of line p
    :param p2: numpy.array; second point of line p
    :param a: numpy.array; first point of line p'
    :param b: numpy.array; second point of line p'
    :return: float
    """
    # todo: add possibility to deside whther on edge point is inside or outside
    return True if _line_side(p1, p2, a, b) >= 0 else False


def is_point_in_polygon(point, polygon):
    """
    test wether point lie in polygon

    :param point: pypex.poly2d.point.Point
    :param polygon: pypex.poly2d.polygon.Polygon
    :return:
    """
    if len(polygon) < 3:
        raise ValueError("invalid polygon shape, expected at least 3 corners polygon")
    return polygon.mplpath.contains_point(point.to_array())


class _Point(object):
    def __init__(self, i, x, y):
        self.i = i
        self.x = x
        self.y = y

    def __key(self):
        return self.x, self.y

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return (other.x == self.x) & (other.y == self.y)


class Point(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __str__(self):
        return "Point [{}, {}]".format(self.x, self.y)

    def __repr__(self):
        return "Point [{}, {}]".format(self.x, self.y)

    def __eq__(self, other):
        return (other.x == self.x) & (other.y == self.y)

    def __add__(self, other):
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return Point(self.x - other.x, self.y - other.y)

    @staticmethod
    def set(points, round_tol=ROUND_PRECISION):
        """
        Naive implementaion `set` like function in python.
        This method relly on tolerance. Points are same up to supplied tolerance.

        :param points: pypex.poly2d.point.Point
        :param round_tol: int
        :return: numpy.arra; list of Point instances
        """
        _points = [_Point(i, round(point.x, round_tol), round(point.y, round_tol)) for i, point in enumerate(points)]
        indices = [_point.i for _point in set(_points)]
        return np.array(points)[indices]

    def is_inside_polygon(self, polygon):
        """
        Test whether current (`self`) Point is inside or outside of polygon
        :param polygon: pypex.poly2d.point.Point
        :return: bool
        """
        return is_point_in_polygon(self, polygon)

    def to_list(self):
        """
        convert Point to list
        :return: list
        """
        return [self.x, self.y]

    def to_array(self):
        """
        Convert Point to numpy.array
        :return: numpy.array
        """
        return np.array(self.to_list())



from .... graphic import data_set_graphics


class Plot(object):
    def __init__(self, instance):
        self.data_set = instance

    def display_observation(self, **plot_kwargs):
        """
        Function that will plot given DataSet observation for visual examination.

        :param plot_kwargs: Dict; arguments passed to matplotlib.pyplot.scatter function (DataSet without errors) or
                                  matplotlib.pyplot.errorbar function (DataSet with errors)
        """
        obs_kwargs = dict()

        obs_kwargs.update({
            'x_data': self.data_set.x_data,
            'y_data': self.data_set.y_data,
            'y_err': self.data_set.y_err,
            'x_unit': self.data_set.x_unit,
            'y_unit': self.data_set.y_unit,
            'plot_kwargs': plot_kwargs
        })
        data_set_graphics.display_observations(**obs_kwargs)



import gc
import numpy as np

from copy import copy
from ... import umpy as up


def initialize_model_container(vertices_map):
    """
    Initializes basic data structure `model` objects that will contain faces sorted by its origin (star or spots)
    and data structure containing spot candidates with its index and center point.
    Structure is based on input `verties_map`.
    Example of return Tuple

    ::

        (<class 'dict'>: {'object': [], 'spots': {0: []}}, <class 'dict'>: {'com': [], 'ix': []})

    :param vertices_map: List or ndarray; map which define refrences of index in
                         given Iterable to object (spot or Star).
                         For more info, see docstring for `incorporate_spots_mesh` method.
    :return: Tuple[Dict, Dict];
    """
    model = {"object": list(), "spots": dict()}
    spot_candidates = {"com": list(), "ix": list()}

    spots_instance_indices = list(set([vertices_map[ix]["enum"] for ix, _ in enumerate(vertices_map)
                                       if vertices_map[ix]["enum"] >= 0]))
    for spot_index in spots_instance_indices:
        model["spots"][spot_index] = list()
    return model, spot_candidates


def split_spots_and_component_faces(star, points, faces, model, spot_candidates, vmap, component_com):
    """
    Function that sorts faces to model data structure by distinguishing if it belongs to star or spots.

    :param star: elisa.base.container.StarContainer;
    :param component_com: float; center of mass of component
    :param points: numpy.array; (N_points * 3) - all points of surface
    :param faces: numpy.array; (N_faces * 3) - all faces of the surface
    :param model: Dict; data structure for faces sorting (more in docstring of method `initialize_model_container`)
    :param spot_candidates: Dict; initialised data structure for spot candidates
    :param vmap: vertice map, for more info, see docstring for `incorporate_spots_mesh` method
    :return: Dict; same as param `model`
    """
    model, spot_candidates = resolve_obvious_spots(points, faces, model, spot_candidates, vmap)
    model = resolve_spot_candidates(star, model, spot_candidates, faces, component_com=component_com)
    # converting lists in model to numpy arrays
    model['object'] = np.array(model['object'])
    for spot_ix in star.spots:
        model['spots'][spot_ix] = np.array(model['spots'][spot_ix])
    return model


def resolve_spot_candidates(star, model, spot_candidates, faces, component_com):
    """
    Resolves spot face candidates by comparing angular distances of face cantres and spot centres.
    In case of multiple layered spots, face is assigned to the top layer.

    :param star: elisa.base.container.StarContainer;
    :param model: Dict; initialised dictionary with placeholders which will describe object with spots as one entity
    :param spot_candidates: Dict; contain indices and center of mass of each
                                  face that is candodate to be a face of spot
    :param faces: ndarray; array of indices which defines faces from points;
    :param component_com: float; center of mass of given component
    :return: Dict; filled model from input of following structure


    ::

        {
            "object": ndarray[[point_idx_i, point_idx_j, point_idx_k], ...],
            "spots": {
                spot_index: ndarray[[point_idx_u, point_idx_v, point_idx_w], ...]
            }
        }
    """
    # checking each candidate one at a time trough all spots
    com = np.array(spot_candidates["com"]) - np.array([component_com, 0.0, 0.0])
    cos_max_angle = {idx: up.cos(_spot.angular_radius) for idx, _spot in star.spots.items()}
    center = {idx: _spot.center - np.array([component_com, 0.0, 0.0]) for idx, _spot in star.spots.items()}
    for idx, _ in enumerate(spot_candidates["com"]):
        spot_idx_to_assign = -1
        simplex_ix = spot_candidates["ix"][idx]
        for spot_ix in star.spots:
            cos_angle_com = up.inner(center[spot_ix], com[idx]) / \
                            (np.linalg.norm(center[spot_ix]) * np.linalg.norm(com[idx]))
            if cos_angle_com > cos_max_angle[spot_ix]:
                spot_idx_to_assign = spot_ix

        if spot_idx_to_assign == -1:
            model["object"].append(np.array(faces[simplex_ix]))
        else:
            model["spots"][spot_idx_to_assign].append(np.array(faces[simplex_ix]))

    gc.collect()
    return model


def resolve_obvious_spots(points, faces, model, spot_candidates, vmap):
    """
    Resolve those Spots/Star faces, where all tree vertices belongs to given object.
    If there are mixed vertices of any face, append it to spot candidate List.

    :param points: numpy.array; array of all points
    :param faces: numpy.array; array of all faces
    :param model: Dict; dictionary which describe object with spots as one entity
    :param spot_candidates:
    :param vmap: vertices map; for more info, see docstring for `incorporate_spots_mesh` method
    :return: Tuple[Dict, Dict]
    """
    for simplex, face_points, ix in list(zip(faces, points[faces], range(faces.shape[0]))):
        # if each point belongs to the same spot, then it is for sure face of that spot
        condition1 = vmap[simplex[0]]["enum"] == vmap[simplex[1]]["enum"] == vmap[simplex[2]]["enum"]
        if condition1:
            if 'spot' == vmap[simplex[0]]["type"]:
                model["spots"][vmap[simplex[0]]["enum"]].append(np.array(simplex))
            else:
                model["object"].append(np.array(simplex))
        else:
            spot_candidates["com"].append(np.average(face_points, axis=0))
            spot_candidates["ix"].append(ix)

    gc.collect()
    return model, spot_candidates


def set_all_surface_centres(star):
    """
    Calculates all surface centres for given body(including spots) and assign to object as `face_centers` property
    """
    star.face_centres = calculate_surface_centres(star.points, star.faces)
    if star.has_spots() and not star.is_flat():
        for spot_index, spot_instance in star.spots.items():
            spot_instance.face_centres = calculate_surface_centres(spot_instance.points, spot_instance.faces)
    return star


def calculate_surface_centres(points, faces):
    """
    Returns centers of every surface face.

    :return: numpy.array;

    ::

        numpy.array([[center_x1, center_y1, center_z1],
                     [center_x2, center_y2, center_z2],
                      ...
                     [center_xn, center_yn, center_zn]])
    """
    return np.average(points[faces], axis=1)


# noinspection PyUnreachableCode
def calculate_normals(points, faces, centres, com):
    """
    Returns outward facing normal unit vector for each face of stellar surface.

    :param points: numpy.array;
    :param faces: numpy.array;
    :param centres: numpy.array;
    :param com: numpy.array;
    :return: numpy.array;

    ::

        numpy.array([[normal_x1, normal_y1, normal_z1],
                     [normal_x2, normal_y2, normal_z2],
                      ...
                     [normal_xn, normal_yn, normal_zn]])
    """
    # vectors defining triangle ABC, a = B - A, b = C - A
    a = points[faces[:, 1]] - points[faces[:, 0]]
    b = points[faces[:, 2]] - points[faces[:, 0]]
    normals = np.cross(a, b)
    normals /= np.linalg.norm(normals, axis=1)[:, None]
    corr_centres = copy(centres) - np.array([com, 0, 0])[None, :]

    # making sure that normals are properly oriented near the axial planes
    sgn = up.sign(np.sum(up.multiply(normals, corr_centres), axis=1))
    return normals * sgn[:, None]


def correct_face_orientation(star_container, com=0):
    """
    Function corrects order if face indices in order to be consistent with outward facing normals.

    :param star_container: elisa.base.container.StarContainer;
    :param com: float; centre of mass x-position
    :return:
    """

    # noinspection PyUnreachableCode,PyUnusedLocal
    def correct_orientation(obj):
        points = getattr(obj, 'points')
        faces = getattr(obj, 'faces')
        centres = getattr(obj, 'face_centres')

        a = points[faces[:, 1]] - points[faces[:, 0]]
        b = points[faces[:, 2]] - points[faces[:, 0]]
        normals = np.cross(a, b)

        corr_centres = copy(centres) - np.array([com, 0, 0])[None, :]

        sgn = up.sign(up.sum(up.multiply(normals, corr_centres), axis=1))
        negative_sgn = sgn < 0
        faces[negative_sgn] = faces[negative_sgn][:, [1, 0, 2]]

    correct_orientation(star_container)
    if star_container.has_spots():
        for spot in star_container.spots.values():
            correct_orientation(spot)

    return star_container


def mirror_triangulation(q_triangles, inverse_point_symmetry_matrix):
    """
    This function enables for the triangulation of symmetrical part of the surface to be mirrored to the rest of
    the surface.

    :param q_triangles: numpy.array; triangles of base symmetry portion of the surface
    :param inverse_point_symmetry_matrix: numpy.array; row-wise - array that map base symmetry portion of the surface
                                                       points to the rest of the surface (quadrants or octants)
    :return: numpy.array; triangulation covering the whole star
    """
    all_triangles = [inv[q_triangles] for inv in inverse_point_symmetry_matrix]
    return np.concatenate(all_triangles, axis=0)


def mirror_face_values(values, face_symmetry_vector):
    """
    Mirroring the array values on symmetrical faces to the rest of the surface.

    :param values: numpy.array; surface parameter values corresponding to the base symmetry part of the surface
    :param face_symmetry_vector: numpy.array; array that remaps `values` to the rest of the surface
    :return: numpy.array; remapped array
    """
    return values[face_symmetry_vector]


def symmetry_face_reduction(values, base_symmetry_faces_number):
    """
    Reducing the surface distribution of surface `values` array to the symmetrical component.

    :param values: numpy.array; surface parameter distribution
    :param base_symmetry_faces_number: int; number of the first `base_symmetry_faces_numbe` from the symmetrical
                                            component
    :return: numpy.array; reduced surface distribution array
    """
    return values[:base_symmetry_faces_number]



import numpy as np

from .. import units, const
from .. base.types import INT, FLOAT
from .. units import DefaultBinarySystemInputUnits
from .. base.transform import SystemProperties, WHEN_FLOAT64, quantity_transform


class BinarySystemProperties(SystemProperties):
    @staticmethod
    def eccentricity(value):
        """
        Transform and validate eccentricity.

        :param value: Union[(numpy.)int, (numpy.)float]
        :return: float;
        """
        if not isinstance(value, (int, INT, float, FLOAT)):
            raise TypeError('Input of variable `eccentricity` is not (numpy.)int or (numpy.)float.')
        if value < 0 or value >= 1:
            raise ValueError('Input of variable `eccentricity` is out of boundaries [0, 1)')
        return np.float64(value)

    @staticmethod
    def argument_of_periastron(value):
        """
        Transform and validate argument of periastron, if unit is not supplied, value in degrees is assumed.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
        :return: float;
        """
        if isinstance(value, (units.Quantity, str)):
            value = units.Quantity(value) if isinstance(value, str) else value
            value = np.float64(value.to(units.DefaultBinarySystemUnits.system.argument_of_periastron))
        elif isinstance(value, WHEN_FLOAT64):
            value = np.float64((value * DefaultBinarySystemInputUnits.system.argument_of_periastron).to(
                units.DefaultBinarySystemUnits.system.argument_of_periastron)
            )
        else:
            raise TypeError('Input of variable `argument_of_periastron` is not (numpy.)int or (numpy.)float '
                            'nor astropy.unit.quantity.Quantity instance.')
        if not 0 <= value <= const.FULL_ARC:
            value %= const.FULL_ARC
        return value

    @staticmethod
    def phase_shift(value):
        """
        Returns phase shift of the primary eclipse minimum with respect to ephemeris
        true_phase is used during calculations, where: true_phase = phase + phase_shift.

        :param value: float;
        :return: float;
        """
        return np.float64(value)

    @staticmethod
    def primary_minimum_time(value):
        """
        Transform and validity check for time of primary minima.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
        :return: float;
        """
        return quantity_transform(value, units.DefaultBinarySystemUnits.system.primary_minimum_time, WHEN_FLOAT64,
                                  units.DefaultBinarySystemInputUnits.system.primary_minimum_time)

    @classmethod
    def t0(cls, value):
        return cls.primary_minimum_time(value)


class RadialVelocityObserverProperties(SystemProperties):
    eccentricity = BinarySystemProperties.eccentricity
    argument_of_periastron = BinarySystemProperties.argument_of_periastron
    period = BinarySystemProperties.period
    gamma = SystemProperties.gamma

    @staticmethod
    def mass_ratio(value):
        """
        Validate mass ratio.

        :param value: float;
        :return: float;
        """
        if not value > 0:
            raise ValueError(f"Invalid value of property `mass_ratio`. Expected > 0, given {value}")
        return FLOAT(value)

    @staticmethod
    def asini(value):
        """
        Transform and validate asini. If value is supplied without unit then default unit is assumed to be solar radii.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
        :return: float;
        """
        if isinstance(value, (units.Quantity, str)):
            value = units.Quantity(value) if isinstance(value, str) else value
            value = np.float64(value.to(units.solRad))
        elif isinstance(value, WHEN_FLOAT64):
            value = np.float64(value)
        else:
            raise TypeError('Input of variable `asini` is not (numpy.)int or (numpy.)float '
                            'nor astropy.unit.quantity.Quantity instance.')
        if value < 0:
            raise ValueError('Value of `asini` cannot be negative.')
        return value



import numpy as np
from .. import const, utils
from .. base.types import INT


def phase_correction(phase, synchronicity):
    """
    Calculate phase correction for mode axis drift.

    :param synchronicity:
    :param phase: float; rotation phase of the star
    :return: float;
    """
    return (synchronicity - 1) * phase * const.FULL_ARC if synchronicity is not np.nan else phase * const.FULL_ARC


def generate_tilt_coordinates(star_container, phase):
    """
    Returns tilt coordinates of pulsation modes.

    :param star_container: StarContainer;
    :param phase: float; rotational orbital phase of the star
    :return: Tuple[float, float];
    """
    phi_corr = phase_correction(phase, star_container.synchronicity) \
        if not star_container.pulsations[0].tidally_locked else 0.0    # we presume that all modes have the same tilt
    phi = star_container.pulsations[0].mode_axis_phi + phi_corr
    theta = star_container.pulsations[0].mode_axis_theta
    return phi, theta


def generate_time_exponential(mode, time):
    """
    Returns time dependent exponential used in `generate_spherical_harmonics`.

    :param mode: PulsationMode; PulsationMode; mode used to generate sph. harmonics
    :param time: float; time at which evaluate spherical harmonics
    :return: float; complex time dependent exponential
    """
    exponent = mode.angular_frequency * time + mode.start_phase
    return np.exp(complex(0, -exponent))


def generate_phase_shift(shift):
    """
    Returns factor that can shift complex displacement by an arbitrary angular phase `shift`.

    :param shift: float; rad
    :return: numpy.complex;
    """
    return np.exp(complex(0, -shift))


def tilt_mode_coordinates(points, phi, theta):
    """
    Function tilts spherical coordinates to desired position described by `phi`, `theta`.

    :param points: numpy.array;
    :param phi: float; azimuthal coordinate of the new polar axis
    :param theta: float; latitude of the new polar axis
    :return: numpy.array;
    """
    if theta != 0 or phi != 0:
        tilted_phi, tilted_theta = utils.rotation_in_spherical(points[:, 1], points[:, 2], phi, theta)
        ret_points = np.column_stack((points[:, 0], tilted_phi, tilted_theta))

        return ret_points
    else:
        return points


def derotate_surface_points(points_to_derotate, phi, theta):
    """
    Derotating surface points  into the base coordinate system after surface displacement for misalligned mode is
    calculated.

    :param points_to_derotate: numpy.array; surface points in tilted spherical coordinates
    :param phi: float; azimuthal tilt of the input coordinate system
    :param theta: float; latitudinal tilt of the input coordinate system
    :return: numpy.array; derotated points in spherical coordinates
    """
    if theta != 0 or phi != 0:
        derot_phi, derot_theta = \
            utils.derotation_in_spherical(points_to_derotate[:, 1],
                                          points_to_derotate[:, 2],
                                          phi, theta)
        return np.column_stack((points_to_derotate[:, 0], derot_phi, derot_theta))
    else:
        return points_to_derotate


def derotate_surface_displacements(velocity, tilted_points, points, axis_phi, axis_theta):
    """
    Transform spherical perturbations from tilted coordinates to system alligned with rotation axis.

    :param velocity: numpy.array; velocity in tilted spherical coordinates
    :param tilted_points: numpy.array; spherical coordinates of surface points (unperturbed) in tilted coordinates
    :param points: numpy.array; unperturbed surface points in spherical coordinates
    :param axis_theta: float;
    :param axis_phi: float;
    :return: numpy.array; perturbations in spherical coordinates aligned with rotation axis
    """
    if axis_theta != 0 or axis_phi != 0:
        pert_phis, pert_thetas = utils.derotation_in_spherical(
            phi=tilted_points[:, 1] + velocity[:, 1],
            theta=tilted_points[:, 2] + velocity[:, 2],
            phi_rotation=axis_phi,
            theta_rotation=axis_theta
        )

        crit_amplitude = const.PI
        d_phi = pert_phis - points[:, 1]
        d_phi[d_phi > crit_amplitude] -= const.FULL_ARC

        d_theta = (pert_thetas - points[:, 2])

        return np.column_stack((velocity[:, 0], d_phi, d_theta))
    else:
        return velocity


def transform_spherical_displacement_to_cartesian(sph_displacement, surf_points, com_x):
    """
    Transforms displacement d_r, d_phi, d_theta into spherical cartesian displacement d_x, d_y, d_z

    :param sph_displacement: numpy.array; [[d_r1, d_phi1, d_theta1], ...]
    :param surf_points: numpy.array; surface points in equilibrium in cartesian coordinates (from container)
    :param com_x: numpy.float; x coordinate of centre of mass, assuming com = [com_x, 0, 0]
    :return: numpy.array; [[d_x1, d_y1, d_z1], ...]
    """
    points = surf_points - np.array([com_x, 0, 0])[None, :]
    r_xy2 = np.sum(np.power(points[:, :-1], 2), axis=1)
    r_xy = np.sqrt(r_xy2)
    r = np.sqrt(r_xy2 + np.power(points[:, 2], 2))
    # z/(x^2+y^2)^0.5
    z_rxy = np.zeros(r_xy.shape)
    non_zero = r_xy != 0
    z_rxy[non_zero] = points[non_zero, 2] / r_xy[non_zero]

    matrix = np.empty((r.shape[0], 3, 3))
    matrix[:, 0, 0], matrix[:, 1, 0], matrix[:, 2, 0] = points[:, 0] / r, -points[:, 1], points[:, 0] * z_rxy
    matrix[:, 0, 1], matrix[:, 1, 1], matrix[:, 2, 1] = points[:, 1] / r,  points[:, 0], points[:, 1] * z_rxy
    matrix[:, 0, 2], matrix[:, 1, 2], matrix[:, 2, 2] = points[:, 2] / r, 0.0, -r_xy

    return np.sum(matrix * sph_displacement[:, :, None], axis=1)


def horizontal_component(displacement, points, treat_poles=False):
    """
    Returns abs value of the horizontal displacement.

    :param displacement: numpy.array; dr, dphi, dtheta
    :param points: numpy.array; r, phi, theta
    :param treat_poles: bool; remove invalid values for faces in contact with pole
    :return:
    """
    # lambda - distance in theta
    # TODO: avoid using sine
    d_lambda = points[:, 0] * np.sin(points[:, 2]) * displacement[:, 1]
    # nu - distance along theta
    d_nu = points[:, 0] * displacement[:, 2]

    distance = np.sqrt(np.power(d_lambda, 2) + np.power(d_nu, 2))
    if treat_poles:
        distance[distance >= 10*distance.mean()] = distance.mean()

    return distance


def pole_neighbours(star):
    """
    Finds indices of both poles and their neighbours.

    :param star: StarContainer
    :return:
    """
    poles = np.array([star.points_spherical[:, 2].argmax(), star.points_spherical[:, 2].argmin()], dtype=INT)
    neighbour_idx = np.empty(2, dtype=INT)
    for ii, pole in enumerate(poles):
        in_face = (pole == star.faces).any(axis=1)
        polar_face = (star.faces[in_face])[0]
        neighbour_idx[ii] = (polar_face[pole != polar_face])[0]

    star.pole_idx = poles
    star.pole_idx_neighbour = neighbour_idx



from elisa.binary_system.surface import mesh
from elisa.binary_system.surface import faces
from elisa.binary_system.surface import gravity
from elisa.binary_system.surface import temperature
from elisa.binary_system.surface import coverage



import numpy as np

from ... const import PI


# _______MCMC_method_______
def s_squared(y_errs, ln_f):
    """
    Calculates error component of the likelihood function derived from the errors.

    :param y_errs: Dict[str, numpy.array]; errors of observables
    :param ln_f: float; error underestimation factor
    :return: Dict[str, float]; s2 parameter
    """
    return {key: np.power(y_errs[key], 2) + np.power(values, 2) * np.exp(2 * ln_f)
            for key, values in y_errs.items()}


def likelihood_fn(y_data, y_errs, synthetic, ln_f):
    """
    Calculates value of likelihood function for observational data being drawn from distribution around synthetic
    model.

    :param y_data: Dict[str, numpy.array]; observables
    :param y_errs: Dict[str, numpy.array]; errors of observables
    :param ln_f: float; marginalization parameters (currently supported single parameter for error penalization)
    :param synthetic: Dict[str, numpy.array]; synthetic observables
    :return: float; likelihood value
    """
    sigma2 = s_squared(y_errs, ln_f)

    lh = - 0.5 * (np.sum(
        [np.sum((np.power((y_data[key] - synthetic[key]), 2) / sigma2[key]) + np.log(2.0 * PI * sigma2[key]))
         for key, value in synthetic.items()])
    )
    return lh


# ________Leat_Squares_method________
def wssr(y_data, y_err, synthetic):
    """
    Return error weighted sum of squared residuals (wssr) for given set of observational and synthetic data.

    :param y_data: Dict[str, numpy.array]; observables
    :param y_err: Dict[str, numpy.array]; errors of observables
    :param synthetic: Dict[str, numpy.array]; synthetic observables
    :return: float; wssr
    """
    return np.sum([np.sum(np.power((synthetic[item] - y_data[item]) / y_err[item], 2))for item in synthetic])





import numpy as np

from .. orbit import orbit
from .. transform import RadialVelocityObserverProperties
from ... logger import getLogger
from ... import (
    umpy as up,
    units as u,
    const
)

logger = getLogger('binary_system.curves.community')


class RadialVelocitySystem(object):
    """
    Standalone class to compute binary components center of mass radial velocity via astro-community parameters.

    Community parameters::

        ``asini`` -- semi_major_axis * sin(inclination)
        ``eccentricit``
        ``argument_of_periastron``
        ``period``
        ``true_anomaly``
        ``gamma``
        ``mass_ratio``

    Input (initialization) parameters

    :param eccentricity: Union[(numpy.)int, (numpy.)float];
    :param argument_of_periastron: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity];
    :param period: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]; Orbital period of binary
                   star system. If unit is not specified, default period unit is assumed (days).
    """

    inclination = const.HALF_PI
    phase_shift = 0.0

    def __init__(self, eccentricity, argument_of_periastron, period, mass_ratio, asini, gamma):
        kwargs = self.transform_input(**dict(
            eccentricity=eccentricity,
            argument_of_periastron=argument_of_periastron,
            period=period,
            mass_ratio=mass_ratio,
            asini=asini,
            gamma=gamma
        ))

        self.eccentricity = np.nan
        self.argument_of_periastron = np.nan
        self.period = np.nan
        self.mass_ratio = np.nan
        self.asini = np.nan
        self.gamma = np.nan
        self.orbit = None
        self.rv_unit = u.dimensionless_unscaled

        self.init_properties(**kwargs)
        self.init_orbit()

    def init_properties(self, **kwargs):
        """
        Setup properties from input.

        :param kwargs: Dict; all supplied input properties
        """
        for key, val in kwargs.items():
            setattr(self, key, val)

    @staticmethod
    def transform_input(**kwargs):
        """
        Transform and validate input kwargs.

        :param kwargs: Dict; model parameters
        :return: Dict;
        """
        return RadialVelocityObserverProperties.transform_input(**kwargs)

    def init_orbit(self):
        """
        Orbit class fro binary like system.
        """
        logger.debug(f"re/initializing orbit in class instance {self.__class__.__name__}")
        orbit_kwargs = {key: getattr(self, key) for key in orbit.Orbit.ALL_KWARGS}
        self.orbit = orbit.Orbit(**orbit_kwargs)

    def radial_velocity(self, **kwargs):
        """
        Method for producing synthetic radial velocity curves in community format (on params `asini` and `q`).

        :param kwargs: Dict;
        :**kwargs options**:
            * :phases: numpy.array; photometric phases used to calculate synthetic radial velocities
            * :position_method: callable; function producing an array of elisa.const.Position tuples that describe
                                          position and orientation of system components at given photometric phase

        """
        phases = kwargs.pop("phases")
        position_method = kwargs.pop("position_method")
        orbital_motion = position_method(phase=phases)

        sma_primary, sma_secondary = self.distance_to_center_of_mass(self.mass_ratio, 1.0)

        period = np.float64((self.period * u.DEFAULT_PERIOD_UNIT).to(u.s))
        asini = np.float64((self.asini * u.solRad).to(u.m))

        sma_primary *= asini
        sma_secondary *= asini

        primary_rv = self._radial_velocity(sma_primary, self.eccentricity, self.argument_of_periastron,
                                           period, orbital_motion[:, 2]) * -1.0

        secondary_rv = self._radial_velocity(sma_secondary, self.eccentricity, self.argument_of_periastron,
                                             period, orbital_motion[:, 2])

        rv_dict = {'primary':  primary_rv + self.gamma, 'secondary': secondary_rv + self.gamma}

        return rv_dict

    @staticmethod
    def distance_to_center_of_mass(q, distance):
        """
        Returns distance from component's centre of mass to the barycentre.

        :param q: float; mass ratio
        :param distance: float; components distance
        :return: Tuple; primary, secondary com distance from barycentre
        """
        com_from_primary = (q * distance) / (1.0 + q)
        return com_from_primary, distance - com_from_primary

    @staticmethod
    def _radial_velocity(asini, eccentricity, argument_of_periastron, period, true_anomaly):
        """
        Compute radial velocity for given paramters.

        :param asini: float;
        :param eccentricity: float;
        :param argument_of_periastron: float;
        :param true_anomaly: Union[float, numpy.array];
        :param period: float
        :return: Union[float, numpy.array];
        """
        a = 2.0 * up.pi * asini
        b = period * up.sqrt(1.0 - up.power(eccentricity, 2))
        c = up.cos(true_anomaly + argument_of_periastron) + (eccentricity * up.cos(argument_of_periastron))
        return - a * c / b

    def get_positions_method(self):
        return self.orbit.orbital_motion

    def compute_rv(self, **kwargs):
        """
        Method for producing synthetic radial velocity curves in community format (on params `asini` and `q`).

        :param kwargs: Dict;
        :**kwargs options**:
            * :phases: numpy.array; photometric phases used to calculate synthetic radial velocities
            * :position_method: callable; function producing an array of elisa.const.Position tuples that describe
                                          position and orientation of system components at given photometric phase

        """
        return self.radial_velocity(**kwargs)

    @staticmethod
    def prepare_json(data):
        """
        Filtering out parameters necessary to initialize RadialVelocitySystem

        :param data: Dict; possibly containing items not valid for initialization of RadialVelocitySystem instance
        :return: Dict; set of argument necessary to initialize RadialVelocitySystem instance
        """
        return dict(
            eccentricity=data['eccentricity'],
            argument_of_periastron=data['argument_of_periastron'],
            period=data['period'],
            mass_ratio=data['mass_ratio'],
            asini=data['asini'],
            gamma=data['gamma']
        )



import numpy as np

from elisa import units, settings, utils
from elisa.utils import transform_values
from elisa.ld import limb_darkening_factor
from elisa.pulse import container_ops, utils as putils


def add_colormap_to_plt_kwargs(*args, **kwargs):
    """
    Returns a colormap that can be passed to surface plot kwargs.

    :param args: tuple;
    :**args options**:
        * **colormap**: str; surface colormap based on given physical parameter, see options below
                             to see available options
        * **star**:  elisa.base.container.StarContainer;
        * **phase**:  float; photometric phase
        * **com_x**:  float; centre of mass
        * **system_scale**:  float; scaling factor of a system
        * **inclination**:  float; inclination of a system
        * **position**:  float; elisa.const.Position

    :**available colormap options**:
        * :'gravity_acceleration': surface distribution of gravity acceleration,
        * :'temperature': surface distribution of the effective temperature,
        * :'velocity': absolute values of surface elements velocities with respect to the observer,
        * :'radial_velocity': radial component of the surface element velocities relative to the observer,
        * :'normal_radiance': surface element radiance perpendicular to the surface element,
        * :'radiance': radiance of the surface element in a direction towards the observer,
        * :'radius': distance of the surface elements from the centre of mass
        * :'horizontal_displacement': distribution of the horizontal component of surface displacement
        * :'horizontal_acceleration': distribution of horizontal component surface acceleration
        * :'v_r_perturbed': radial component of the pulsation velocity (perpendicular towards
        * :'v_horizontal_perturbed': horizontal component of the pulsation  velocity

    :param kwargs: Dict;
    :**kwargs options**:
        * **scale**: str; `log` or `linear`
        * **unit**: astropy.units.Unit;
        * **subtract_equilibrium**: bool; if True; equilibrium values are subtracted from surface colormap

    :return: numpy.array;
    """
    colorbar_fn = {
        'radius': r_cmap,
        'horizontal_displacement': horizonatal_displacement_cmap,
        'gravity_acceleration': g_cmap,
        'temperature': t_cmap,
        'velocity': v_cmap,
        'horizontal_acceleration': horizontal_g_pert_cmap,
        'v_r_perturbed': v_rad_pert_cmap,
        'v_horizontal_perturbed': v_horizontal_pert_cmap,
        'radial_velocity': v_rad_cmap,
        'normal_radiance': norm_radiance_cmap,
        'radiance': radiance_cmap,
    }

    colormap, star, phase, com_x, model_scale, inclination, position = args
    scale = kwargs.get('scale', 'linear')
    unit = kwargs.get('unit', 'default')
    subtract_equilibrium = kwargs.get('subtract_equilibrium', False)

    retval = None
    if colormap is None:
        return retval

    if colormap not in colorbar_fn.keys():
        raise KeyError(f'Unknown `colormap` argument {colormap}. Options: {colorbar_fn.keys()}')
    if subtract_equilibrium:
        if not star.has_pulsations():
            raise ZeroDivisionError('You are trying to display surface colormap with `subtract_equilibrium`=True but '
                                    'surface of the star does not oscillate.')

    args = (scale, unit, subtract_equilibrium, model_scale, inclination, position)
    retval = colorbar_fn[colormap](star, *args)

    return retval


def r_cmap(star, *args):
    """
    Returning the radius of the points as a colormap.

    :param star: elisa.base.container.StarContainer;
    :param args: Tuple;
    :**args contents**:
        * :scale: str; log or linear
        * :unit: astropy.units.Unit; unit for the colormap
        * :subtract_equilibrium: bool; if true, return only perturbation from equilibrium state
        * :model_scale: float; scale of the system

    :return: numpy.array;
    """
    scale, unit, subtract_equilibrium, model_scale = args[:4]
    if not subtract_equilibrium:
        points = star.points - star.com[None, :]
        value = utils.cartesian_to_spherical(points)[:, 0]
    else:
        args = (star, 0.0, False, True, True)
        perturbation = container_ops.position_perturbation(*args)
        value = perturbation[:, 0]

    value = value[star.faces].mean(axis=1) * model_scale
    unt = units.DISTANCE_UNIT if unit == 'default' else unit
    value = transform_values(value, units.DISTANCE_UNIT, unt)
    return to_log(value, scale)


def horizonatal_displacement_cmap(star, *args):
    """
    Returning the horizontal component of the surface element displacement as a colormap.

    :param star: elisa.base.container.StarContainer;
    :param args: Tuple;
    :**args contents**:
        * :scale: str; log or linear
        * :unit: astropy.units.Unit; unit for the colormap
        * :subtract_equilibrium: bool; if true, return only perturbation from equilibrium state
        * :model_scale: float; scale of the system

    :return: numpy.array;
    """
    scale, unit, subtract_equilibrium, model_scale = args[:4]
    if not subtract_equilibrium and not star.has_pulsations():
        raise ValueError('Horizontal displacement colormap is relevant only for stars with pulsations.')
    args = (star, 0.0, False, True, True)
    perturbation = container_ops.position_perturbation(*args)

    value = putils.horizontal_component(perturbation, star.points_spherical)
    value = value[star.faces].mean(axis=1) * model_scale
    unt = units.DISTANCE_UNIT if unit == 'default' else unit
    value = transform_values(value, units.DISTANCE_UNIT, unt)
    return to_log(value, scale)


def v_cmap(star, *args):
    """
    Returning surface element speed colormap.

    :param star: elisa.base.container.StarContainer;
    :param args: Tuple;
    :**args contents**:
        * :scale: str; log or linear
        * :unit: astropy.units.Unit; unit for the colormap
        * :subtract_equilibrium: bool; if true, return only perturbation from equilibrium state
        * :model_scale: float; scale of the system

    :return: numpy.array;
    """
    scale, unit, subtract_equilibrium, model_scale = args[:4]
    args = (star, model_scale, False, True, False)
    velocities = container_ops.velocity_perturbation(*args) if subtract_equilibrium else getattr(star, 'velocities')
    velocities = np.linalg.norm(velocities, axis=1)
    unt = units.m / units.s if unit == 'default' else unit
    value = transform_values(velocities, units.VELOCITY_UNIT, unt)
    return to_log(value, scale)


def v_rad_cmap(star, *args):
    """
    Returning radial velocity colormap (with respect to the observer).

    :param star: elisa.base.container.StarContainer;
    :**args contents**:
        * :scale: str; log or linear
        * :unit: astropy.units.Unit; unit for the colormap
        * :subtract_equilibrium: bool; if true, return only perturbation from equilibrium state
        * :model_scale: float; scale of the system
        * :inclination: float;
        * :position: elisa.const.Position;

    :return: numpy.array;
    """
    scale, unit, subtract_equilibrium, model_scale, inclination, position = args
    if subtract_equilibrium:
        args = (star, model_scale, False, True, False)
        velocities = container_ops.velocity_perturbation(*args)
        velocities = utils.rotate_item(velocities, position, inclination)
    else:
        velocities = getattr(star, 'velocities')
    velocities = velocities[:, 0]
    unt = units.m / units.s if unit == 'default' else unit
    value = transform_values(velocities, units.VELOCITY_UNIT, unt)
    if scale in ['log', 'logarithmic']:
        raise Warning("`log` scale is not allowed for radial velocity colormap.")
    return value


def v_rad_pert_cmap(star, *args):
    """
    Returning radial component of the velocity perturbation.

    :param star: elisa.base.container.StarContainer;
    :param args: Tuple;
    :**args contents**:
        * :scale: str; log or linear
        * :unit: astropy.units.Unit; unit for the colormap
        * :subtract_equilibrium: bool; if true, return only perturbation from equilibrium state
        * :model_scale: float; scale of the system

    :return: numpy.array;
    """
    scale, unit, subtract_equilibrium, model_scale = args[:4]
    if not subtract_equilibrium and not star.has_pulsations():
        raise ValueError('`v_r_perturbed` is relevant only for stars with pulsations.')
    args = (star, model_scale, False, True, True)
    velocities = container_ops.velocity_perturbation(*args)[:, 0]
    unt = units.m / units.s if unit == 'default' else unit
    value = transform_values(velocities, units.VELOCITY_UNIT, unt)
    return to_log(value, scale)


def v_horizontal_pert_cmap(star, *args):
    """
    Returning horizontal component of the velocity perturbation.

    :param star: elisa.base.container.StarContainer;
    :param args: Tuple;
    :**args contents**:
        * :scale: str; log or linear
        * :unit: astropy.units.Unit; unit for the colormap
        * :subtract_equilibrium: bool; if true, return only perturbation from equilibrium state
        * :model_scale: float; scale of the system

    :return: numpy.array;
    """
    scale, unit, subtract_equilibrium, model_scale = args[:4]
    if not subtract_equilibrium and not star.has_pulsations():
        raise ValueError('`v_horizontal_perturbed` colormap is relevant only for stars with pulsations.')
    args = (star, model_scale, False, True, True)
    velocities = container_ops.velocity_perturbation(*args)
    face_centres_sph = star.points_spherical[star.faces].mean(axis=1)
    velocities = putils.horizontal_component(velocities, face_centres_sph, treat_poles=True) * model_scale
    unt = units.m / units.s if unit == 'default' else unit
    value = transform_values(velocities, units.VELOCITY_UNIT, unt)
    return to_log(value, scale)


def g_cmap(star, *args):
    """
    Returning gravity acceleration colormap.

    :param star: elisa.base.container.StarContainer;
    :param args: Tuple;
    :**args contents**:
        * :scale: str; log or linear
        * :unit: astropy.units.Unit; unit for the colormap
        * :subtract_equilibrium: bool; if true, return only perturbation from equilibrium state
        * :model_scale: float; scale of the system

    :return: numpy.array;
    """
    scale, unit, subtract_equilibrium, model_scale = args[:4]
    if subtract_equilibrium:
        if scale in ['log', 'logarithmic']:
            raise ValueError('Logarithmic scale is not permitted with the `subtract_equilibrium` = True.')
        args = (star, model_scale, False, True, True)
        g = container_ops.gravity_acc_perturbation(*args)[:, 0]
    else:
        log_g = getattr(star, 'log_g')
        g = np.power(10, log_g)
    value = transform_values(g, units.ACCELERATION_UNIT, unit)
    return to_log(value, scale)


def horizontal_g_pert_cmap(star, *args):
    """
    Returning horizontal component of the acceleration perturbation.

    :param star: elisa.base.container.StarContainer;
    :param args: Tuple;
    :**args contents**:
        * :scale: str; log or linear
        * :unit: astropy.units.Unit; unit for the colormap
        * :subtract_equilibrium: bool; if true, return only perturbation from equilibrium state
        * :model_scale: float; scale of the system

    :return: numpy.array;
    """
    scale, unit, subtract_equilibrium, model_scale = args[:4]
    if not subtract_equilibrium and not star.has_pulsations():
        raise ValueError('`horizontal_acceleration` colormap is relevant only for stars with pulsations.')
    args = (star, model_scale, False, True, True)
    acceleration = container_ops.gravity_acc_perturbation(*args)
    face_centres_sph = star.points_spherical[star.faces].mean(axis=1)
    acceleration = putils.horizontal_component(acceleration, face_centres_sph, treat_poles=True) * model_scale
    unt = units.ACCELERATION_UNIT if unit == 'default' else unit
    value = transform_values(acceleration, units.ACCELERATION_UNIT, unt)
    return to_log(value, scale)


def t_cmap(star, *args):
    """
    Returning temperature colormap.

    :param star: elisa.base.container.StarContainer;
    :param args: Tuple;
    :**args contents**:
        * :scale: str; log or linear
        * :unit: astropy.units.Unit; unit for the colormap
        * :subtract_equilibrium: bool; if true, return only perturbation from equilibrium state
        * :model_scale: float; scale of the system

    :return: numpy.array;
    """
    scale, unit, subtract_equilibrium, model_scale = args[:4]
    argss = (star, model_scale, False, True)
    temperatures = container_ops.temp_perturbation(*argss) if subtract_equilibrium else getattr(star, 'temperatures')
    value = transform_values(temperatures, units.DefaultStarInputUnits.t_eff, unit)
    return to_log(value, scale)


def norm_radiance_cmap(star, *args):
    """
    Returning radiance in the direction of surface normal vector.

    :param star: elisa.base.container.StarContainer;
    :param args: Tuple;
    :**args contents**:
        * :scale: str; log or linear
        * :unit: astropy.units.Unit; unit for the colormap
        * :subtract_equilibrium: bool; if true, return only perturbation from equilibrium state

    :return: numpy.array;
    """
    scale, unit, subtract_equilibrium = args[:3]
    normal_radiance = getattr(star, 'normal_radiance')['bolometric']
    value = transform_values(normal_radiance, units.RADIANCE_UNIT, unit)
    return to_log(value, scale)


def radiance_cmap(star, *args):
    """
    Returning radiance in the direction of the observer.

    :param star: elisa.base.container.StarContainer;
    :param args: Tuple;
    :**args contents**:
        * :scale: str; log or linear
        * :unit: astropy.units.Unit; unit for the colormap
        * :subtract_equilibrium: bool; if true, return only perturbation from equilibrium state

    :return: numpy.array;
    """
    scale, unit = args[:2]
    normal_radiance = getattr(star, 'normal_radiance')['bolometric']
    los_cosines = getattr(star, 'los_cosines')
    indices = getattr(star, 'indices')
    ld_cfs = getattr(star, 'ld_cfs')['bolometric'][indices]
    ld_cors = limb_darkening_factor(coefficients=ld_cfs,
                                    limb_darkening_law=settings.LIMB_DARKENING_LAW,
                                    cos_theta=los_cosines[indices])
    retval = np.zeros(normal_radiance.shape)
    retval[indices] = normal_radiance[indices] * los_cosines[indices] * ld_cors

    value = transform_values(retval, units.RADIANCE_UNIT, unit)
    if scale in ['log', 'logarithmic']:
        raise Warning("`log` scale is not allowed for radiance colormap.")
    return value


def to_log(value, scale):
    """
    Function transforms to logarithmic scale.

    :param value: value: Union[float, numpy.array]; input values
    :param scale: str; `log`, `logarithmic`, `linear`
    :return: numpy.array;
    """
    return np.log10(value) if scale in ['log', 'logarithmic'] else value



from abc import ABCMeta, abstractmethod
from multiprocessing import Pool
from typing import Dict

import emcee
import numpy as np
from scipy import interpolate
from scipy.stats.distributions import norm

from . shared import check_for_boundary_surface_potentials
from . mixins import MCMCMixin
from . shared import (
    lc_r_squared, rv_r_squared,
    AbstractLCFit, AbstractRVFit, AbstractFit
)

from .. import RVData, LCData
from .. models import lc as lc_model
from .. models import rv as rv_model
from .. models import cost_fns
from .. params import parameters
from .. params.conf import NUISANCE_PARSER, PARAM_PARSER
from .. tools.utils import time_layer_resolver

from ... observer.utils import normalize_light_curve
from ... base.error import ElisaError
from ... import settings
from ... graphic.mcmc_graphics import Plot
from ... logger import getPersistentLogger
from ... binary_system.system import BinarySystem
from ... binary_system.curves.community import RadialVelocitySystem

logger = getPersistentLogger('analytics.binary_fit.mcmc')


class MCMCFit(AbstractFit, MCMCMixin, metaclass=ABCMeta):
    """
    General framework for MCMC sampling of binary systems.
    """
    def __init__(self):
        self.plot = Plot()
        self.last_sampler = emcee.EnsembleSampler
        self.last_normalization = dict()
        self.flat_chain_path = ''
        self.eval_counter = 0
        self._last_known_lhood = -np.finfo(float).max * np.finfo(float).eps
        self.sigmas = None

    @staticmethod
    def ln_prior(xn, x0, sigmas):
        """
        Logarithmic value of prior (uniform, normal or combined).

        :param xn: numpy.array; current state of the sampler (normalized values of variable parameters)
        :param x0: numpy.array; mean (expected) values of normalized parameters with normal prior distribution
        :param sigmas: numpy.array; normalized standard deviations of normal prior distributions
        :return: numpy.array; sum of logarithms of prior distribution functions
        """
        retval = np.empty(sigmas.shape)

        nan_mask = np.isnan(sigmas)
        uni_prior = np.all(np.bitwise_and(np.greater_equal(xn[nan_mask], 0.0),
                                          np.less_equal(xn[nan_mask], 1.0))).astype(float)
        retval[nan_mask] = -np.inf if uni_prior == 0 else np.log(uni_prior)

        retval[~nan_mask] = np.log(norm().pdf(((xn[~nan_mask] - x0[~nan_mask]) / sigmas[~nan_mask]))) \
            if np.logical_and(xn[~nan_mask] >= 0, xn[~nan_mask] <= 1.0).all() else -np.inf
        return np.sum(retval)

    @abstractmethod
    def likelihood(self, xn):
        """
        Likelihood function depending on the type of the optimization.

        :param xn: numpy.array; current state of the sampler (normalized values of variable parameters)
        :return: float;
        """
        pass

    def likelihood_fn(self, synthetic, ln_f):
        """
        Calculates value of likelihood function for observational data being drawn from distribution around synthetic
        model.

        :param ln_f: float; marginalization parameters (currently supported single parameter for error penalization)
        :param synthetic: Dict; {'dataset_name': numpy.array, }
        :return: float; likelihood value
        """
        lh = cost_fns.likelihood_fn(self.y_data, self.y_err, synthetic, ln_f)
        self._last_known_lhood = lh if lh < self._last_known_lhood else self._last_known_lhood
        return lh

    def ln_probability(self, xn):
        """
        Resulting probability distribution made of likelihood and prior distribution.

        :param xn: numpy.array; current state of the sampler (normalized values of variable parameters)
        :return: float; likelihood
        """
        prior = self.ln_prior(xn, self.norm_init_vector, self.sigmas)
        if prior == -np.inf:
            return -np.inf
        try:
            likelihood = prior + self.likelihood(xn)
        except (ElisaError, ValueError) as e:
            if not settings.SUPPRESS_WARNINGS:
                logger.warning(f'mcmc hit invalid parameters, exception: {str(e)}')
            return self._last_known_lhood * 1e3
        return likelihood

    def normalized_sigma(self, vector):
        """
        Assigns normalized standard deviation for each variable parameter to attribute `sigma`. If sigma is not supplied for the
        parameter, np.nan is used instead.

        :param vector: List; normalized starting vector
        :return: numpy.array;
        """
        sigmas = np.array([val.sigma if val.sigma is not None else np.nan for val in self.fitable.values()])
        perturbed = np.array(self.initial_vector) + sigmas
        perturbed_norm = parameters.vector_normalizer(perturbed, self.fitable.keys(), self.normalization)
        return np.array(perturbed_norm) - vector

    def _fit(self, nwalkers, ndim, nsteps, nsteps_burn_in, p0=None, progress=False, save=False, fit_id=None):
        """
        General MCMC sampling function for an inverse problem. Implementing sampler from the emcee package.

        :param nwalkers: int; The number of walkers in the ensemble. Minimum is 2 * number of free parameters.
        :param ndim: int; number of free variables
        :param nsteps: int; The number of steps to run.
        :param nsteps_burn_in: int; number of steps for mcmc to explore parameters
        :param p0: numpy.array; initial priors for mcmc
        :param progress: bool; display the progress bar of the sampling
        :param save: bool; if true, the MCMC flat chain will be stored
        :param fit_id: str; id or location (ending with .json) which identifies fit file (if not specified, current
                            datetime is used)
        :return: emcee.EnsembleSampler;
        """
        self.norm_init_vector = np.array(parameters.vector_normalizer(self.initial_vector, self.fitable.keys(),
                                                                      self.normalization))

        self.sigmas = self.normalized_sigma(self.norm_init_vector)
        # sigmas = np.array([val.sigma if val.sigma is not None else np.nan for val in self.fitable.values()])
        # args = (np.array(self.initial_vector) + sigmas, self.fitable.keys(), self.normalization)
        # perturbed_norm = parameters.vector_normalizer(*args)
        # self.sigmas = np.array(perturbed_norm) - self.norm_init_vector

        p0 = self.generate_initial_states(p0, nwalkers, ndim, x0_vector=self.norm_init_vector)

        logger.info('starting mcmc')
        kwargs = dict(nwalkers=nwalkers, ndim=ndim, log_prob_fn=self.ln_probability)
        if settings.NUMBER_OF_MCMC_PROCESSES > 1:
            with Pool(processes=settings.NUMBER_OF_MCMC_PROCESSES) as pool:
                logger.info('starting parallel mcmc')
                sampler = emcee.EnsembleSampler(pool=pool, **kwargs)
                self.worker(sampler, p0, nsteps, nsteps_burn_in, save=save, fit_id=fit_id, fitable=self.fitable,
                            normalization=self.normalization, progress=progress)
        else:
            logger.info('starting singlecore mcmc')
            sampler = emcee.EnsembleSampler(**kwargs)
            self.worker(sampler, p0, nsteps, nsteps_burn_in, save=save, fit_id=fit_id, fitable=self.fitable,
                        normalization=self.normalization, progress=progress)

        self.last_sampler = sampler
        self.last_normalization = self.normalization

        if save:
            self.flat_chain_path = self.save_flat_chain(sampler.get_chain(flat=True), self.fitable,
                                                        self.normalization, fit_id=fit_id)

        return sampler

    @staticmethod
    def generate_initial_states(initial_state, nwalkers, ndim, x0_vector=None):
        """
        Function transforms user initial state to normalized format suitable for our MCMC chain, where all vales are in
        interval (0, 1).

        :param initial_state: numpy.ndarray; initial state matrix before normalization
        :param nwalkers: int; The number of walkers in the ensemble. Minimum is 2 * number of free parameters.
        :param ndim: int; number of free variables
        :param x0_vector: np.array; normalized vector of free parameters
        :return: initial_state: numpy.ndarray; initial state matrix after normalization
        """
        if initial_state is None:
            retval = np.random.uniform(0.0, 1.0, (nwalkers, ndim))
            retval[0] = x0_vector if x0_vector is not None else retval[0]
            return retval
        else:
            if initial_state.shape != (nwalkers, ndim):
                raise ValueError(f'Your initial values for sampler do not satisfy required shape ({nwalkers}, {ndim}). '
                                 f'Shape of your initial state matrix is {initial_state.shape}')
            initial_state[initial_state < 0] = 0.0
            initial_state[initial_state > 1] = 1.0
            return initial_state


class LightCurveFit(MCMCFit, AbstractLCFit):
    """
    LC fit class implementing an MCMC method.
    """
    MORPHOLOGY = None

    def likelihood(self, xn):
        """
        Likelihood function for given set of model variables.
        Best fit is 0.0, worst is -inf.

        :param xn: Iterable[float]; vector of optimized free parameters
        :return: float; likelihood
        """
        diff = 1.0 / self.interp_treshold
        xn = parameters.vector_renormalizer(xn, self.fitable.keys(), self.normalization)
        kwargs = parameters.prepare_properties_set(xn, self.fitable.keys(), self.constrained, self.fixed)
        phases, kwargs = time_layer_resolver(self.x_data_reduced, pop=False, **kwargs)

        fit_xs = np.linspace(np.min(phases) - diff, np.max(phases) + diff, num=self.interp_treshold + 2) \
            if np.shape(phases)[0] > self.interp_treshold else phases
        args = fit_xs, self.discretization, self.observer
        synthetic = lc_model.synthetic_binary(*args, **kwargs)
        synthetic, _ = normalize_light_curve(synthetic, kind='average')

        if np.shape(fit_xs) != np.shape(phases):
            synthetic = {
                band: interpolate.interp1d(fit_xs, curve, kind='cubic')(phases[self.x_data_reducer[band]])
                for band, curve in synthetic.items()
            }

        ln_f_key = f"{NUISANCE_PARSER}{PARAM_PARSER}ln_f"
        ln_f = parameters.prepare_nuisance_properties_set(xn, self.fitable, self.fixed)[ln_f_key]

        return self.likelihood_fn(synthetic, ln_f)

    def fit(self, data: Dict[str, LCData], x0: parameters.BinaryInitialParameters, discretization=5.0, nwalkers=None,
            nsteps=1000, initial_state=None, burn_in=None, percentiles=None, interp_treshold=None, progress=False,
            save=True, fit_id=None, samples="uniform"):
        """
        Fit method using Markov Chain Monte Carlo.
        Once simulation is done, following values are stored and can be used for further evaluation::

            self.last_sampler: emcee.EnsembleSampler
            self.last_normalization: Dict; normalization map used during fitting
            self.flat_chain_path: str; filename of last stored flatten emcee `sampler` with metadata

        Based on https://emcee.readthedocs.io/en/stable/.

        :param data: elisa.analytics.dataset.base.LCData;
        :param x0: List[Dict]; initial state (metadata included)
        :param discretization: float; discretization of objects
        :param interp_treshold: int; Above this total number of datapoints, light curve will be interpolated
                                     using model containing `interp_treshold` equidistant points per epoch
        :param nwalkers: int; The number of walkers in the ensemble. Minimum is 2 * number of free parameters.
        :param nsteps: int; The number of steps to run.
        :param initial_state: numpy.array; initial priors for MCMC
        :param burn_in: int; number of steps for MCMC to explore parameters
        :param progress: bool; display the progress bar of the sampling
        :param percentiles: List; [percentile for left side error estimation, percentile of the centre,
                                   percentile for right side error estimation]
        :param save: bool; whether to store the chain or not
        :param fit_id: str; id which identifies fit file (if not specified, current dateime is used)
        :param samples: Union[str, List]; `uniform`, `adaptive` or list with phases in (0, 1) interval
        :return: Dict; optimized model parameters in flattened form
        """
        burn_in = int(nsteps / 10) if burn_in is None else burn_in
        self.set_up(x0, data, passband=data.keys(), discretization=discretization, morphology=self.MORPHOLOGY,
                    interp_treshold=settings.MAX_CURVE_DATA_POINTS if interp_treshold is None else interp_treshold,
                    observer_system_cls=BinarySystem, samples=samples)

        ndim = len(self.initial_vector)
        nwalkers = 2 * len(self.initial_vector) if nwalkers is None else nwalkers
        self.mcmc_nwalkers_vs_ndim_validity_check(nwalkers, ndim)

        sampler = self._fit(nwalkers, ndim, nsteps, burn_in, initial_state, progress, save, fit_id)

        # extracting fit results from MCMC sampler
        flat_chain = sampler.get_chain(flat=True)
        result_dict = MCMCMixin.resolve_mcmc_result(flat_chain, self.fitable, self.normalization, percentiles)

        result_dict.update({lbl: {"value": val.value, 'fixed': True, 'unit': val.to_dict()['unit']}
                            for lbl, val in self.fixed.items()})
        result_dict = self.eval_constrained_results(result_dict, self.constrained)

        r_squared_args = (self.x_data_reduced, self.y_data, self.observer.passband, discretization,
                          self.x_data_reducer, 1.0 / self.interp_treshold, self.interp_treshold,
                          self.observer.system_cls)

        r_dict = {key: value['value'] for key, value in result_dict.items()}
        r_dict = parameters.extend_json_with_atm_params(
            r_dict,
            atmosphere_model=self.atmosphere_model,
            limb_darkening_coefficients=self.limb_darkening_coefficients
        )

        r_squared_result = lc_r_squared(lc_model.synthetic_binary, *r_squared_args, **r_dict)
        result_dict["r_squared"] = {'value': r_squared_result, "unit": None}

        result_dict = check_for_boundary_surface_potentials(result_dict, LightCurveFit.MORPHOLOGY)

        setattr(self, 'flat_result', result_dict)
        return parameters.serialize_result(result_dict)


class OvercontactLightCurveFit(LightCurveFit):
    """
    MCMC fitting implementation for light curves of over-contact binaries.
    It keeps eye on values of potentials - keep it same for primary and secondary component.
    """
    MORPHOLOGY = 'over-contact'


class DetachedLightCurveFit(LightCurveFit):
    """
    MCMC fitting implementation for light curves of detached binaries.
    """
    MORPHOLOGY = 'detached'


class CentralRadialVelocity(MCMCFit, AbstractRVFit):
    """
    RV fit class implementing an MCMC method using kinematic method.
    """
    def likelihood(self, xn):
        """
        Likelihood function for given set of model parameters.
        Best is 0.0, worst is -inf.

        :param xn: Iterable[float]; vector of parameters we are looking for
        :return: float;
        """
        xn = parameters.vector_renormalizer(xn, self.fitable.keys(), self.normalization)
        kwargs = parameters.prepare_properties_set(xn, self.fitable.keys(), self.constrained, self.fixed)
        synthetic = rv_model.central_rv_synthetic(*(self.x_data_reduced, self.observer), **kwargs)
        synthetic = {comp: rv[self.x_data_reducer[comp]] for comp, rv in synthetic.items()}

        ln_f_key = f"{NUISANCE_PARSER}{PARAM_PARSER}ln_f"
        ln_f = parameters.prepare_nuisance_properties_set(xn, self.fitable, self.fixed)[ln_f_key]
        lhood = self.likelihood_fn(synthetic, ln_f)

        self.eval_counter += 1
        logger.debug(f'eval counter = {self.eval_counter}, likehood = {lhood}')
        return lhood

    def fit(self, data: Dict[str, RVData], x0: parameters.BinaryInitialParameters, nwalkers=None, nsteps=1000,
            initial_state=None, burn_in=None, percentiles=None, progress=False, save=True, fit_id=None):
        """
        Fit method using Markov Chain Monte Carlo.
        Once simulation is done, following values are stored and can be used for further evaluation::
        sampler = self._fit(x0_vector, self.labels, nwalkers, ndim, nsteps, nsteps_burn_in, p0)

        # extracting fit results from MCMC sampler::

            Based on https://emcee.readthedocs.io/en/stable/.

        :param save: bool; whether to store chain or not
        :param fit_id: str; id which identifies fit file (if not specified, current dateime is used)
        :param data: elisa.analytics.dataset.base.RVData;
        :param x0: elisa.analytics.params.parameters.BinaryInitialParameters;
        :param nwalkers: int; The number of walkers in the ensemble. Minimum is 2 * number of free parameters.
        :param nsteps: int; The number of steps to run.
        :param initial_state: numpy.array; initial priors for mcmc
        :param burn_in: int; numer of steps for mcmc to explore parameters
        :param percentiles: List[int]; [percentile for left side error estimation, percentile of the centre,
                                       percentile for right side error estimation]
        :param progress: bool; visualize progress of the sampling
        :return: Dict; fit results
        """
        burn_in = int(nsteps / 10) if burn_in is None else burn_in
        self.set_up(x0, data, observer_system_cls=RadialVelocitySystem)

        ndim = len(self.initial_vector)
        nwalkers = 2 * len(self.initial_vector) if nwalkers is None else nwalkers
        self.mcmc_nwalkers_vs_ndim_validity_check(nwalkers, ndim)

        sampler = self._fit(nwalkers, ndim, nsteps, burn_in, initial_state, progress, save, fit_id)

        # extracting fit results from MCMC sampler
        flat_chain = sampler.get_chain(flat=True)
        result_dict = MCMCMixin.resolve_mcmc_result(flat_chain, self.fitable, self.normalization, percentiles)

        result_dict.update({lbl: {
            'value': val.value,
            'fixed': True,
            'unit': val.to_dict()['unit']
        } for lbl, val in self.fixed.items()})
        result_dict = self.eval_constrained_results(result_dict, self.constrained)

        r_squared_args = self.x_data_reduced, self.y_data, self.x_data_reducer, self.observer.system_cls
        r_dict = {key: value['value'] for key, value in result_dict.items()}

        r_squared_result = rv_r_squared(rv_model.central_rv_synthetic, *r_squared_args, **r_dict)
        result_dict["r_squared"] = {'value': r_squared_result, "unit": None}

        setattr(self, 'flat_result', result_dict)
        return parameters.serialize_result(result_dict)



import numpy as np

from copy import deepcopy
from . surface import (
    mesh,
    faces,
    gravity,
    temperature,
    pulsations
)
from .. logger import getLogger
from .. import utils
from .. base.container import (
    StarContainer,
    PositionContainer
)

logger = getLogger("binary_system.container")


class OrbitalPositionContainer(PositionContainer):
    """
    Object for handling models of a binary system at given orbital position.
    Use functions OrbitalPositionContainer.from_binary_system or
    BinarySystem.build_container to correctly initialize this container.
    """
    def __init__(self, primary: StarContainer, secondary: StarContainer, position, **properties):
        """
        Initialization of the OrbitalPositionContainer using StarContainers

        :param primary: elisa.base.container.StarContainer;
        :param secondary: elisa.base.container.StarContainer;
        :param position: elisa.const.Position;
        :param properties: Dict; serialized binary system parameters
        """
        super().__init__(position=position)
        self._components = ['primary', 'secondary']
        self.primary = primary
        self.secondary = secondary

        # placeholder (set in loop below)
        self.period = np.nan

        for key, val in properties.items():
            setattr(self, key, val)

        # calculating a time that elapsed since t0
        self.time = self.set_time()

        # setting centre of mass
        self.set_com(self.position)

    def set_on_position_params(self, position, primary_potential=None, secondary_potential=None):
        """
        Defining the orbital position and orientation of the binary system container with respect
        to the observer.

        :param position: elisa.const.Position;
        :param primary_potential: float; corrected surface potential of the primary corresponding to the `position`
        :param secondary_potential: float; corrected surface potential of the secondary corresponding to the `position`
        :return: OrbitalPositionContainer; container with set position and surface potentials
        """
        setattr(self, "position", position)
        self.set_com(position)
        if not utils.is_empty(primary_potential):
            setattr(self.primary, "surface_potential", primary_potential)
        if not utils.is_empty(secondary_potential):
            setattr(self.secondary, "surface_potential", secondary_potential)
        return self

    def set_com(self, position):
        """
        Calculate and set centre of masses of both components in frame of reference centered to the primary component.

        :param position: elisa.const.Position;
        """
        setattr(self.primary, 'com', np.array([0, 0, 0]))
        setattr(self.secondary, 'com', np.array([position.distance, 0, 0]))
        self.rotate_property(self.primary, 'com')
        self.rotate_property(self.secondary, 'com')

    def set_time(self):
        """
        Calculating elapsed from `primary_minimum_time` for the instance of OrbitalPositionContainer.

        :return: float; time in seconds corresponding to the container
        """
        setattr(self, 'time', 86400 * self.period * self.position.phase)
        return getattr(self, 'time')

    @classmethod
    def from_binary_system(cls, binary_system, position):
        """
        Construct an OrbitalPositionContainer based on the instance of BinarySystem and orbital position.

        :param binary_system: elisa.binary_system.system.BinarySystem;
        :param position: elisa.const.Position; named tuple containing information concerning orbital
                                               position of the binary components and their orientation
                                               in space with respect to the observer.
        :return: OrbitalPositionContainer;
        """
        radii = binary_system.calculate_components_radii(position.distance)
        primary = StarContainer.from_star_instance(binary_system.primary)
        secondary = StarContainer.from_star_instance(binary_system.secondary)
        primary.assign_radii(radii['primary'])
        secondary.assign_radii(radii['secondary'])
        return cls(primary, secondary, position, **binary_system.properties_serializer())

    def copy(self):
        """Returns a deep copy of the OrbitalPositionContainer."""
        return deepcopy(self)

    def has_spots(self):
        """Returns True if at least one component contains spots."""
        return self.primary.has_spots() or self.secondary.has_spots()

    def has_pulsations(self):
        """Returns True if at least one component contains pulsations."""
        return self.primary.has_pulsations() or self.secondary.has_pulsations()

    def build(self, components_distance=None, component="all", build_pulsations=True, **kwargs):
        """
        Main method to build binary star system from parameters given on init of BinaryStar.

        called following methods::

            - build_mesh
            - build_faces
            - build_velocities
            - build_surface_gravity
            - build_faces_orientation
            - correct_mesh
            - build_surface_areas
            - build_temperature_distribution

        :param component: str; `primary` or `secondary`
        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :param build_pulsations: bool; if True, only equilibrium model is build
        :return: OrbitalPositionContainer;
        """

        components_distance = self._components_distance(components_distance)
        self.build_mesh(components_distance, component)
        self.build_from_points(components_distance, component)

        # flatt it: from this point we do not require separated information about spots
        self.flat_it()
        if build_pulsations:
            self.build_pulsations(components_distance=components_distance, component=component)
        return self

    def build_pulsations(self, components_distance=None, component="all"):
        """
        Incorporating user-defined pulsation modes into the model.

        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :param component: str; 'primary`, `secondary` or `all`
        """
        self.build_harmonics(components_distance=components_distance, component=component)
        self.build_perturbations(components_distance=components_distance, component=component)

    def build_from_points(self, components_distance=None, component="all"):
        """
        Function is used on container to build container on which only bulid_mesh was performed. Function builds the
        rest of geometries

        Order of methods::

            - build_faces
            - build_velocities
            - build_surface_gravity
            - build_faces_orientation
            - correct_mesh
            - build_surface_areas

        and temperatures.

        Order of methods::

            - build_temperature_distribution

        :param component: str; 'primary`, `secondary` or `all`
        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :return: OrbitalPositionContainer; self
        """
        self.build_faces_and_kinematic_quantities(components_distance, component)
        self.build_temperature_distribution(components_distance, component)
        return self

    def build_faces_and_kinematic_quantities(self, components_distance=None, component="all"):
        """
        Function is used on container to build container on which only
        build_mesh was performed. Function builds the rest except for
        build_temperature_distribution.

        :param component: str; 'primary`, `secondary` or `all`
        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :return: OrbitalPositionContainer;
        """
        components_distance = self._components_distance(components_distance)
        self.build_faces(components_distance, component)
        self.build_velocities(components_distance, component)
        self.build_surface_gravity(components_distance, component)
        self.build_faces_orientation(components_distance, component)
        self.correct_mesh(components_distance, component)
        self.build_surface_areas(component)

        return self

    def build_mesh(self, components_distance=None, component="all"):
        """
        Build surface points for primary or/and secondary component. In case of spots,
        the spot point mesh is incorporated into the model.

        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :param component: str; 'primary`, `secondary` or `all`
        :return: OrbitalPositionContainer; container updated with point mesh
        """
        components_distance = self._components_distance(components_distance)
        return mesh.build_mesh(self, components_distance, component)

    def correct_mesh(self, components_distance=None, component="all"):
        """
        Correcting the underestimation of the surface due to the discretization.

        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :param component: str; 'primary`, `secondary` or `all`
        :return: OrbitalPositionContainer; container updated with corrected point mesh
        """
        return mesh.correct_mesh(self, components_distance=components_distance, component=component)

    def rebuild_symmetric_detached_mesh(self, components_distance=None, component="all"):
        """
        Rebuild a mesh of a symmetrical surface using old mesh to provide azimuths for the new.
        This conserved number of points and faces and saves computational time during a surface
        recalculation in case of similar orbital positions.

        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :param component: str; 'primary`, `secondary` or `all`
        :return: OrbitalPositionContainer; container updated with rebuilt point mesh
        """
        components_distance = self._components_distance(components_distance)
        return mesh.rebuild_symmetric_detached_mesh(self, components_distance, component)

    def build_faces(self, components_distance=None, component="all"):
        """
        Function creates faces of the star surface for given components. Faces are
        evaluated upon points that have to be in this time already calculated.

        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :param component: str; 'primary`, `secondary` or `all`
        :return: OrbitalPositionContainer; container updated with faces
        """
        components_distance = self._components_distance(components_distance)
        return faces.build_faces(self, components_distance, component)

    def build_velocities(self, components_distance=None, component='all'):
        """
        Function calculates velocity vector for each face relative to the system's centre of mass.

        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :param component: str; 'primary`, `secondary` or `all`
        :return: OrbitalPositionContainer; container updated with face velocities
        """
        components_distance = self._components_distance(components_distance)
        return faces.build_velocities(self, components_distance, component)

    def build_surface_areas(self, component="all"):
        """
        Compute surface are of all faces (spots included).

        :param component: str; 'primary`, `secondary` or `all`
        :return: OrbitalPositionContainer; container updated with face (triangle) areas
        """
        return faces.compute_all_surface_areas(self, component)

    def build_faces_orientation(self, components_distance=None, component="all"):
        """
        Compute face orientation (normals) for each face.

        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :param component: str; 'primary`, `secondary` or `all`
        :return: OrbitalPositionContainer; container updated with correct normal vector orientation for each face
        """
        components_distance = self._components_distance(components_distance)
        return faces.build_faces_orientation(self, components_distance, component)

    def build_surface_gravity(self, components_distance=None, component="all"):
        """
        Function calculates gravity potential gradient magnitude (surface gravity) for each face.
        Value assigned to face is calculated as a mean of surface gravity values calculated in corners of given face.

        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :param component: str; 'primary`, `secondary` or `all`
        :return: OrbitalPositionContainer; container updated with surface gravity distribution
        """
        components_distance = self._components_distance(components_distance)
        return gravity.build_surface_gravity(self, components_distance, component)

    def build_temperature_distribution(self, components_distance=None, component="all", do_pulsations=False,):
        """
        Function calculates temperature distribution across all surface faces.

        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :param component: str; 'primary`, `secondary` or `all`
        :param do_pulsations: bool; if True, pulsations are incorporated into the system
        :return: OrbitalPositionContainer; container updated with surface temperature distribution
        """
        components_distance = self._components_distance(components_distance)
        return temperature.build_temperature_distribution(self, components_distance, component)

    def build_harmonics(self, component, components_distance):
        """
        Adds pre-calculated spherical harmonics values for each pulsation mode.

        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :param component: str; 'primary`, `secondary` or `all`
        :return: OrbitalPositionContainer; container updated with pre-calculated spherical harmonics
        """
        return pulsations.build_harmonics(self, component, components_distance)

    def build_perturbations(self, component, components_distance):
        """
        Incorporating perturbations of surface quantities into the PositionContainer.

        :param components_distance: Union[None, float]; distance of components is SMA units.
                                                        If None, OrbitalPositionContainer.position.distance is used
        :param component: str; 'primary`, `secondary` or `all`
        :return: OrbitalPositionContainer; container with introduced pulsations
        """
        return pulsations.build_perturbations(self, component, components_distance)

    def _components_distance(self, components_distance):
        return components_distance if components_distance is not None else self.position.distance



from ... import units as u
from elisa.base.transform import (
    TransformProperties,
    quantity_transform,
    WHEN_FLOAT64
)


class OrbitProperties(TransformProperties):
    @staticmethod
    def rotational_period(value):
        """
        Transform and validate rotational period of single star system.
        If unit is not specified, default period unit is assumed.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
        :return: float
        """
        return quantity_transform(value, u.DefaultSingleSystemUnits.system.rotation_period, WHEN_FLOAT64)

    @staticmethod
    def inclination(value):
        """
        If unitless values is supplied, default units suppose to be radians.

        :param value: Union[float, astropy.units.Quantity]
        :return: float
        """
        return quantity_transform(value, u.DefaultSingleSystemUnits.system.inclination, WHEN_FLOAT64)



from . import (
    lc_point,
    c_router
)
from ... binary_system import dynamic

# main wrapper over lc computation


def compute_circular_synchronous_lightcurve(binary, **kwargs):
    """
    Compute light curve for synchronous circular binary system.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float
        * ** position_method** * - function definition; to evaluate orbital positions
        * ** phases ** * - numpy.array
    :return: Dict[str, numpy.array];
    """
    initial_system = c_router.prep_initial_system(binary)

    band_labels = [*kwargs["passband"].keys()]
    phases = kwargs.pop("phases")
    unique_phase_interval, reverse_phase_map = dynamic.phase_crv_symmetry(initial_system, phases)

    _args = (binary, initial_system, unique_phase_interval, lc_point.compute_lc_on_pos, band_labels)
    band_curves = c_router.produce_circular_sync_curves(*_args, **kwargs)
    band_curves = {band: band_curves[band][reverse_phase_map] for band in band_curves}

    return band_curves


def compute_circular_spotty_asynchronous_lightcurve(binary, **kwargs):
    """
    Function returns light curve of asynchronous systems with circular orbits and spots.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** - float
        * ** right_bandwidth ** - float
        * ** atlas ** - str

    :return: Dict; fluxes for each filter
    """
    lc_labels = [*kwargs["passband"].keys()]
    return c_router.produce_circular_spotty_async_curves(binary, lc_point.compute_lc_on_pos, lc_labels, **kwargs)


def compute_circular_pulsating_lightcurve(binary, **kwargs):
    """
    Function returns light curve of pulsating binary systems with circular orbits.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** - float
        * ** right_bandwidth ** - float
        * ** atlas ** - str
        * ** phases ** * - numpy.array

    :return: Dict; fluxes for each filter
    """
    initial_system = c_router.prep_initial_system(binary, **dict(build_pulsations=False))
    band_labels = list(kwargs["passband"].keys())
    _args = (binary, initial_system, kwargs.pop("phases"), lc_point.compute_lc_on_pos, band_labels)
    return c_router.produce_circular_pulsating_curves(*_args, **kwargs)


def compute_eccentric_lightcurve_no_spots(binary, **kwargs):
    """
    General function for generating light curves of binaries with eccentric orbit and no spots.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param kwargs:  Dict;
    :**kwargs options**:
        * ** passband ** - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** - float
        * ** right_bandwidth ** - float
        * ** atlas ** - str

    :return: Dict; fluxes for each filter
    """
    lc_labels = [*kwargs["passband"].keys()]
    return c_router.produce_ecc_curves_no_spots(binary, lc_point.compute_lc_on_pos, lc_labels, **kwargs)


def compute_eccentric_spotty_lightcurve(binary, **kwargs):
    """
    Function returns light curve of systems with eccentric orbits and spots.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** - float
        * ** right_bandwidth ** - float
        * ** atlas ** - str

    :return: Dict; dictionary of fluxes for each filter
    """
    lc_labels = list(kwargs["passband"].keys())
    return c_router.produce_ecc_curves_with_spots(binary, lc_point.compute_lc_on_pos, lc_labels, **kwargs)



from ... import umpy as np
from elisa.base.curves import utils as crv_utils


def _calculate_rv_point(star):
    """
    Calculates point on the rv curve for given component.

    :param star: elisa.base.container.StarContainer; star container with all necessary parameters pre-calculated
    :return: Union[numpy.float, numpy.nan];
    """
    indices = getattr(star, 'indices')
    velocities = getattr(star, 'velocities')[indices]
    fluxes = crv_utils.calculate_surface_element_fluxes('rv_band', star)
    return np.sum(velocities[:, 0] * fluxes) / np.sum(fluxes) \
        if np.sum(fluxes) != 0 else np.NaN


def compute_rv_at_pos(velocities, pos_idx, crv_labels, system):
    """
    Calculates rv points for given orbital position.

    :param velocities: Dict; {str; component : numpy.array; rvs, ...}
    :param pos_idx: int; position in `band_curves` to which calculated lc points will be assigned
    :param crv_labels: List; list of components for which to calculate rvs
    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :return: Dict; updated {str; passband : numpy.array; rvs, ...}
    """
    for component in crv_labels:
        velocities[component][pos_idx] = _calculate_rv_point(getattr(system, component))
    return velocities



"""
Converting input parameters to default internal units.
"""


import numpy as np

from .. params import conf
from ... import units as u
from ... base.transform import (
    TransformProperties,
    WHEN_FLOAT64,
    StarProperties, SpotProperties
)
from ... binary_system.transform import (
    BinarySystemProperties,
    RadialVelocityObserverProperties
)
from ... pulse.transform import PulsationModeProperties


def angular(value):
    """
    Transform all angular units to ELISa's default internal angular unit.

    :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
    :return: float;
    """
    if isinstance(value, u.Quantity):
        value = np.float64(value.to(conf.DEFAULT_FLOAT_ANGULAR_UNIT))
    elif isinstance(value, WHEN_FLOAT64):
        value = np.float64(value)
    else:
        raise TypeError('Input of variable is not (numpy.)int or (numpy.)float '
                        'nor astropy.unit.quantity.Quantity instance.')
    return value


class BinaryInitialProperties(TransformProperties):
    """
    Module that handles conversion of binary system parameters units.
    """
    @staticmethod
    def semi_major_axis(value):
        if isinstance(value, u.Quantity):
            value = np.float64(value.to(u.solRad))
        elif isinstance(value, WHEN_FLOAT64):
            value = np.float64(value)
        else:
            raise TypeError('Input of variable `semi_major_axis` is not (numpy.)int or (numpy.)float '
                            'nor astropy.unit.quantity.Quantity instance.')
        if value < 0:
            raise ValueError('Value of `semi_major_axis` cannot be negative.')
        return value

    @staticmethod
    def mass(value):
        if isinstance(value, u.Quantity):
            value = np.float64(value.to(conf.DEFAULT_FLOAT_MASS_UNIT))
        elif isinstance(value, WHEN_FLOAT64):
            value = np.float64(value)
        else:
            raise TypeError('User input is not (numpy.)int or (numpy.)float '
                            'nor astropy.unit.quantity.Quantity instance.')
        if value <= 0:
            raise ValueError("Invalid mass, use value > 0!")
        return value

    eccentricity = BinarySystemProperties.eccentricity
    argument_of_periastron = angular
    inclination = angular
    gamma = BinarySystemProperties.gamma
    period = BinarySystemProperties.period
    mass_ratio = RadialVelocityObserverProperties.mass_ratio
    asini = RadialVelocityObserverProperties.asini
    additional_light = BinarySystemProperties.additional_light
    primary_minimum_time = BinarySystemProperties.primary_minimum_time


class StarInitialProperties(StarProperties):
    """
    Module that handles conversion of component's parameters units.
    """
    mass = BinaryInitialProperties.mass


class SpotInitialProperties(SpotProperties):
    """
    Module that handles conversion of spot's parameters units.
    """
    latitude = angular
    longitude = angular
    angular_radius = angular


class NuisanceInitialProperties(TransformProperties):
    """
    Module that handles unit conversion of nuisance fit parameters.
    """
    ln_f = lambda x: x


class PulsationModeInitialProperties(PulsationModeProperties):
    """
    Module that handles unit conversion of pulsation mode parameters.
    """
    mode_axis_theta = angular
    mode_axis_phi = angular



import os
import json

from elisa import settings


def load_standard(system):
    """
    Loads zero points for magnitude calculations.

    :param system: str; available: `vega`, 'ab', 'st'
    :return: dict; containing zero point for each filter
    """
    file = os.path.join(settings.DATA_PATH, "zero_points", f'{system.lower()}.json')
    with open(file, 'r') as fl:
        standard_fluxes = json.load(fl)

    return standard_fluxes



def eval_args_for_magnitude_gradient(star_container):
    """
    returns either all of the surface pints (not spots) in case of the surface with spots or just its symmetrical part
    in case of clean surface

    :param star_container: StarContainer;
    :return: Tuple;
    """
    if star_container.symmetry_test():
        points = star_container.symmetry_points()
        faces = star_container.symmetry_faces(star_container.faces)
    else:
        points, faces = star_container.points, star_container.faces
    return points, faces



import numpy as np
import json
from copy import copy
from ... import umpy as up, settings
from ... logger import getLogger


logger = getLogger("base.surface.temperature")


def calculate_effective_temperatures(star_container, gradient_magnitudes):
    """
    Calculates effective temperatures for given gradient magnitudes.
    If None is given, star surface t_effs are calculated.

    :param star_container: elisa.base.container.StarContainer;
    :param gradient_magnitudes: numpy.array;
    :return: numpy.array;
    """

    t_eff_polar = calculate_polar_effective_temperature(star_container)
    t_eff = t_eff_polar * up.power(gradient_magnitudes / star_container.polar_potential_gradient_magnitude,
                                   0.25 * star_container.gravity_darkening)
    return star_container.mirror_face_values(t_eff) if star_container.symmetry_test() else t_eff


def calculate_polar_effective_temperature(star_container):
    """
    Returns polar effective temperature.

    :param star_container: elisa.base.container.StarContainer;
    :return:
    """
    areas = copy(star_container.areas)
    potential_gradient_magnitudes = star_container.potential_gradient_magnitudes
    if star_container.has_spots():
        for idx, spot in star_container.spots.items():
            areas = up.concatenate((areas, spot.areas), axis=0)
            potential_gradient_magnitudes = \
                up.concatenate((potential_gradient_magnitudes, spot.potential_gradient_magnitudes), axis=0)

    return star_container.t_eff * up.power(np.sum(areas) /
                                           np.sum(areas * up.power(
                                               potential_gradient_magnitudes /
                                               star_container.polar_potential_gradient_magnitude,
                                               star_container.gravity_darkening)),
                                           0.25)


def renormalize_temperatures(star):
    """
    In case of spot presence, renormalize temperatures to match the effective temperature of the star,
    since spots disrupt the original effective temperature of Star.
    """
    # no need to calculate surfaces they had to be calculated already, otherwise there is nothing to renormalize
    total_surface = np.sum(star.areas)
    if star.has_spots():
        for spot_index, spot in star.spots.items():
            total_surface += np.sum(spot.areas)
    desired_flux_value = total_surface * up.power(star.t_eff, 4)

    current_flux = np.sum(star.areas * up.power(star.temperatures, 4))
    if star.spots:
        for spot_index, spot in star.spots.items():
            current_flux += np.sum(spot.areas * up.power(spot.temperatures, 4))

    coefficient = up.power(desired_flux_value / current_flux, 0.25)
    logger.debug(f'surface temperature map renormalized by a factor {coefficient}')
    star.temperatures *= coefficient
    if star.spots:
        for spot_index, spot in star.spots.items():
            spot.temperatures *= coefficient


def interpolate_bolometric_gravity_darkening(temperature):
    """
    Quick beta interpolator based on Figure 2 in Claret 2003, A&A 406, 623–628.

    :param temperature: float
    :return: float; interpolated beta
    """
    with open(settings.PATH_TO_BETA, 'r') as f:
        data = json.load(f)
    interp_temps = data['x']
    interp_betas = data['y']

    if temperature <= 0:
        raise ValueError('Negative temperature of the star encountered.')

    return np.interp(np.log10(temperature), interp_temps, interp_betas)
